# Copyright 2018 NTT DATA.  All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from masakari.objects import fields


create_compute_host_payload = {
    'type': 'object',
    'properties': {
        'host_status': {
            'enum': fields.HostStatusType.ALL,
            'type': 'string'},
        'cluster_status': {
            'enum': fields.ClusterStatusType.ALL,
            'type': 'string'},
        'event': {
            'enum': fields.EventType.ALL,
            'type': 'string'},
    },
    'required': ['event'],
    'additionalProperties': False
}

create_process_payload = {
    'type': 'object',
    'properties': {
        'process_name': {
            'type': 'string',
            'minLength': 1,
            'maxLength': 4096},
        'event': {
            'enum': fields.EventType.ALL,
            'type': 'string'},
        },
    'required': ['process_name', 'event'],
    'additionalProperties': False
}

create_vm_payload = {
    'type': 'object',
    'properties': {
        'instance_uuid': {
            'type': 'string',
            'format': 'uuid'},
        'vir_domain_event': {
            'type': 'string',
            'minLength': 1,
            'maxLength': 255},
        'event': {
            'type': 'string',
            'minLength': 1,
            'maxLength': 255},
    },
    'required': ['instance_uuid', 'vir_domain_event', 'event'],
    'additionalProperties': False
}



#    Copyright (c) 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_utils import timeutils
from oslo_versionedobjects import fields

from masakari import objects
from masakari.tests import uuidsentinel

NOW = timeutils.utcnow().replace(microsecond=0)


def fake_db_notification(**updates):
    db_notification = {
        "type": "VM",
        "id": 1,
        "payload":
            {'event': 'STOPPED', 'host_status': 'NORMAL',
             'cluster_status': 'ONLINE'
             },
        "source_host_uuid": uuidsentinel.fake_host,
        "generated_time": NOW,
        "status": "running",
        "notification_uuid": uuidsentinel.fake_notification,
        "created_at": NOW,
        "updated_at": None,
        "deleted_at": None,
        "deleted": 0
    }

    for name, field in objects.Notification.fields.items():
        if name in db_notification:
            continue
        if field.nullable:
            db_notification[name] = None
        elif field.default != fields.UnspecifiedDefault:
            db_notification[name] = field.default
        else:
            raise Exception('fake_db_notification needs help with %s' % name)

    if updates:
        db_notification.update(updates)

    return db_notification


def fake_notification_obj(context, **updates):
    return objects.Notification(**updates)



# Copyright 2016 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from migrate import ForeignKeyConstraint, UniqueConstraint
from sqlalchemy import Column, MetaData, Table, Index
from sqlalchemy import Integer, DateTime, String, Boolean, Text


def define_hosts_table(meta):
    failover_segments = Table('failover_segments', meta, autoload=True)
    hosts = Table('hosts',
                  meta,
                  Column('created_at', DateTime, nullable=False),
                  Column('updated_at', DateTime),
                  Column('deleted_at', DateTime),
                  Column('deleted', Integer),
                  Column('id', Integer, primary_key=True,
                         nullable=False),
                  Column('uuid', String(36), nullable=False),
                  Column('name', String(255), nullable=False),
                  Column('reserved', Boolean, default=False),
                  Column('type', String(255), nullable=False),
                  Column('control_attributes', Text, nullable=False),
                  Column('failover_segment_id', String(36), nullable=False),
                  Column('on_maintenance', Boolean, default=False),
                  UniqueConstraint('failover_segment_id', 'name', 'deleted',
                                   name='uniq_host0name0deleted'),
                  UniqueConstraint('uuid', name='uniq_host0uuid'),
                  ForeignKeyConstraint(columns=['failover_segment_id'],
                                       refcolumns=[failover_segments.c.uuid],
                                       name='fk_failover_segments_uuid'),
                  Index('hosts_type_idx', 'type'),
                  mysql_engine='InnoDB',
                  mysql_charset='utf8',
                  extend_existing=True)

    return hosts


def upgrade(migrate_engine):
    meta = MetaData()
    meta.bind = migrate_engine
    table = define_hosts_table(meta)
    table.create()



# Copyright 2016 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Defines interface for DB access.

Functions in this module are imported into the masakari.db namespace.
Call these functions from masakari.db namespace, not the masakari.db.api
namespace.
"""

from oslo_db import concurrency

import masakari.conf

CONF = masakari.conf.CONF

_BACKEND_MAPPING = {'sqlalchemy': 'masakari.db.sqlalchemy.api'}

IMPL = concurrency.TpoolDbapiWrapper(CONF, backend_mapping=_BACKEND_MAPPING)

# The maximum value a signed INT type may have
MAX_INT = 0x7FFFFFFF


def get_engine():
    """Returns database engine"""
    return IMPL.get_engine()


def failover_segment_get_all_by_filters(
        context, filters=None, sort_keys=None, sort_dirs=None,
        limit=None, marker=None):
    """Get all failover segments that match all filters.

    :param context: context to query under
    :param filters: filters for the query in the form of key/value
    :param sort_keys: list of attributes by which results should be sorted,
                    paired with corresponding item in sort_dirs
    :param sort_dirs: list of directions in which results should be sorted,
                    paired with corresponding item in sort_keys
    :param limit: maximum number of items to return
    :param marker: the last item of the previous page, used to determine the
                  next page of results to return

    :returns: list of dictionary-like objects containing all failover segments
    """
    return IMPL.failover_segment_get_all_by_filters(context, filters=filters,
                                                    sort_keys=sort_keys,
                                                    sort_dirs=sort_dirs,
                                                    limit=limit,
                                                    marker=marker)


def failover_segment_get_by_id(context, segment_id):
    """Get failover segment by id.

    :param context: context to query under
    :param segment_id: id of failover segment

    :returns: dictionary-like object containing failover segment

    :raises exception.FailoverSegmentNotFound if failover segment with given ID
            doesn't exist.
    """
    return IMPL.failover_segment_get_by_id(context, segment_id)


def failover_segment_get_by_uuid(context, segment_uuid):
    """Get failover segment by uuid.

    :param context: context to query under
    :param segment_uuid: uuid of failover segment

    :returns: dictionary-like object containing failover segment

    :raises exception.FailoverSegmentNotFound if failover segment with given
            'segment_uuid' doesn't exist.
    """
    return IMPL.failover_segment_get_by_uuid(context, segment_uuid)


def failover_segment_get_by_name(context, name):
    """Get failover segment by name

    :param context: context: context to query under
    :param name: name of failover segment

    :returns: dictionary-like object containing failover segment

    :raises exception.FailoverSegmentNotFoundByName if failover segment with
            given 'name' doesn't exist.
    """
    return IMPL.failover_segment_get_by_name(context, name)


def failover_segment_create(context, values):
    """Insert failover segment to database.

    :param context: context to query under
    :param values: dictionary of failover segment attributes to create

    :returns: dictionary-like object containing created failover segment

    :raises exception.FailoverSegmentExists if failover segment with given name
            already exist.
    """
    return IMPL.failover_segment_create(context, values)


def failover_segment_update(context, segment_uuid, values):
    """Update failover segment by uuid.

    :param context: context to query under
    :param segment_uuid: uuid of segment to be updated
    :param values: dictionary of values to be updated

    :returns: dictionary-like object containing updated failover segment

    :raises exception.FailoverSegmentNotFound if failover segment with given
            'segment_uuid' doesn't exist.
            exception.FailoverSegmentExists if failover segment with given name
            already exist.
    """
    return IMPL.failover_segment_update(context, segment_uuid, values)


def failover_segment_delete(context, segment_uuid):
    """Delete the failover segment.

    :param context: context to query under
    :param segment_uuid: uuid of segment to be deleted

    :raises exception.FailoverSegmentNotFound if failover segment with
            'segment_uuid' doesn't exist.
    """
    return IMPL.failover_segment_delete(context, segment_uuid)


def is_failover_segment_under_recovery(context, failover_segment_id,
                                       filters=None):
    """Checks whether failover segment is used for processing any notification

    :param context: context to query under
    :param failover_segment_id: uuid of segment
    :param filters: dictionary of filters; values that are lists, tuples,
                    sets, or frozensets cause an 'IN' test to be performed,
                    while exact matching ('==' operator) is used for other
                    values.

    :returns: Returns True if any of the host belonging to a failover segment
              is being used for processing any notifications which are in
              new, error or running status otherwise it will return False.
    """
    return IMPL.is_failover_segment_under_recovery(
        context, failover_segment_id, filters=filters)


# db apis for host


def host_get_all_by_filters(
        context, filters=None, sort_keys=None, sort_dirs=None,
        limit=None, marker=None):
    """Get all hosts that match all filters.

    :param context: context to query under
    :param filters: filters for the query in the form of key/value
    :param sort_keys: list of attributes by which results should be sorted,
                     paired with corresponding item in sort_dirs
    :param sort_dirs: list of directions in which results should be sorted,
                     paired with corresponding item in sort_keys
    :param limit: maximum number of items to return
    :param marker: the last item of the previous page, used to determine the
                   next page of results to return

    :returns: list of dictionary-like objects containing all hosts
    """
    return IMPL.host_get_all_by_filters(context, filters=filters,
                                        sort_keys=sort_keys,
                                        sort_dirs=sort_dirs, limit=limit,
                                        marker=marker)


def host_get_by_uuid(context, host_uuid, segment_uuid=None):
    """Get host information by uuid.

    :param context: context to query under
    :param host_uuid: uuid of host
    :param segment_uuid: uuid of failover_segment

    :returns: dictionary-like object containing host

    :raises: exception.HostNotFound if host with 'host_uuid' doesn't exist
    """
    return IMPL.host_get_by_uuid(context, host_uuid, segment_uuid=segment_uuid)


def host_get_by_id(context, host_id):
    """Get host information by id.

    :param context: context to query under
    :param host_id: id of host

    :returns: dictionary-like object containing host

    :raises: exception.HostNotFound if host with given ID doesn't exist
    """
    return IMPL.host_get_by_id(context, host_id)


def host_get_by_name(context, name):
    """Get host information by name.

    :param context: context to query under
    :param name: name of host

    :returns: dictionary-like object containing host

    :raises: exception.HostNotFoundByName if host with given 'name' doesn't
             exist
    """
    return IMPL.host_get_by_name(context, name)


def host_create(context, values):
    """Create a host.

    :param context: context to query under
    :param values: dictionary of host attributes to create

    :returns: dictionary-like object containing created host
    """
    return IMPL.host_create(context, values)


def host_update(context, host_uuid, values):
    """Update host information in the database.

    :param context: context to query under
    :param host_uuid: uuid of host to be updated
    :param values: dictionary of host attributes to be updated

    :returns: dictionary-like object containing updated host

    :raises: exception.HostNotFound if host with 'host_uuid' doesn't exist
             exception.HostExists if host with given 'name' already exist
    """
    return IMPL.host_update(context, host_uuid, values)


def host_delete(context, host_uuid):
    """Delete the host.

    :param context: context to query under
    :param host_uuid: uuid of host to be deleted

    :raises: exception.HostNotFound if host with 'host_uuid' doesn't exist
    """
    return IMPL.host_delete(context, host_uuid)


# notification related db apis


def notifications_get_all_by_filters(
        context, filters=None, sort_keys=None, sort_dirs=None,
        limit=None, marker=None):
    """Get all notifications that match all filters.

    :param context: context to query under
    :param filters: filters for the query in the form of key/value
    :param sort_keys: list of attributes by which results should be sorted,
                     paired with corresponding item in sort_dirs
    :param sort_dirs: list of directions in which results should be sorted,
                     paired with corresponding item in sort_keys
    :param limit: maximum number of items to return
    :param marker: the last item of the previous page, used to determine the
                   next page of results to return

    :returns: list of dictionary-like objects containing all notifications
    """
    return IMPL.notifications_get_all_by_filters(context, filters=filters,
                                                 sort_keys=sort_keys,
                                                 sort_dirs=sort_dirs,
                                                 limit=limit,
                                                 marker=marker)


def notification_get_by_uuid(context, notification_uuid):
    """Get notification information by uuid.

    :param context: context to query under
    :param notification_uuid: uuid of notification

    :returns: dictionary-like object containing notification

    :raises: exception.NotificationNotFound if notification with given
             'notification_uuid' doesn't exist
    """
    return IMPL.notification_get_by_uuid(context, notification_uuid)


def notification_get_by_id(context, notification_id):
    """Get notification information by id.

    :param context: context to query under
    :param notification_id: id of notification

    :returns: dictionary-like object containing notification

    :raises: exception.NotificationNotFound if notification with given ID
             doesn't exist
    """
    return IMPL.notification_get_by_id(context, notification_id)


def notification_create(context, values):
    """Create a notification.

    :param context: context to query under
    :param values: dictionary of notification attributes to create

    :returns: dictionary-like object containing created notification
    """
    return IMPL.notification_create(context, values)


def notification_update(context, notification_uuid, values):
    """Update notification information in the database.

    :param context: context to query under
    :param notification_uuid: uuid of notification to be updated
    :param values: dictionary of notification attributes to be updated

    :returns: dictionary-like object containing updated notification

    :raises: exception.NotificationNotFound if notification with given
             'notification_uuid' doesn't exist
    """
    return IMPL.notification_update(context, notification_uuid, values)


def notification_delete(context, notification_uuid):
    """Delete the notification.

    :param context: context to query under
    :param notification_uuid: uuid of notification to be deleted

    :raises: exception.NotificationNotFound if notification with given
             'notification_uuid' doesn't exist
    """
    return IMPL.notification_delete(context, notification_uuid)


def purge_deleted_rows(context, age_in_days, max_rows):
    """Purge the soft deleted rows.

    :param context: context to query under
    :param age_in_days: Purge deleted rows older than age in days
    :param max_rows: Limit number of records to delete
    """
    return IMPL.purge_deleted_rows(context, age_in_days, max_rows)



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.
"""Example Module A for testing utils.monkey_patch()."""


def example_function_a():
    return 'Example function'


class ExampleClassA(object):
    def example_method(self):
        return 'Example method'

    def example_method_add(self, arg1, arg2):
        return arg1 + arg2



# Copyright 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import contextlib
import os

from oslo_log import log as logging
from oslo_utils import timeutils
from stevedore import named
# For more information please visit: https://wiki.openstack.org/wiki/TaskFlow
import taskflow.engines
from taskflow import exceptions
from taskflow import formatters
from taskflow.listeners import base
from taskflow.listeners import logging as logging_listener
from taskflow.persistence import backends
from taskflow.persistence import models
from taskflow import task

import masakari.conf
from masakari import exception

CONF = masakari.conf.CONF
PERSISTENCE_BACKEND = CONF.taskflow.connection
LOG = logging.getLogger(__name__)


class MasakariTask(task.Task):
    """The root task class for all masakari tasks.

    It automatically names the given task using the module and class that
    implement the given task as the task name.
    """

    def __init__(self, context, novaclient, **kwargs):
        requires = kwargs.get('requires')
        rebind = kwargs.get('rebind')
        provides = kwargs.get('provides')
        super(MasakariTask, self).__init__(self.__class__.__name__,
                                           requires=requires,
                                           rebind=rebind,
                                           provides=provides)
        self.context = context
        self.novaclient = novaclient
        self.progress = []

    def update_details(self, progress_data, progress=0.0):
        progress_details = {
            'timestamp': str(timeutils.utcnow()),
            'progress': progress,
            'message': progress_data
        }

        self.progress.append(progress_details)
        self._notifier.notify('update_progress', {'progress': progress,
                                                  "progress_details":
                                                      self.progress})


class SpecialFormatter(formatters.FailureFormatter):

    # Exception is an excepted case, don't include traceback in log if fails.
    _NO_TRACE_EXCEPTIONS = (exception.SkipInstanceRecoveryException,
                            exception.SkipHostRecoveryException)

    def __init__(self, engine):
        super(SpecialFormatter, self).__init__(engine)

    def format(self, fail, atom_matcher):
        if fail.check(*self._NO_TRACE_EXCEPTIONS) is not None:
            exc_info = None
            exc_details = '%s%s' % (os.linesep, fail.pformat(traceback=False))
            return (exc_info, exc_details)
        else:
            return super(SpecialFormatter, self).format(fail, atom_matcher)


class DynamicLogListener(logging_listener.DynamicLoggingListener):
    """This is used to attach to taskflow engines while they are running.

    It provides a bunch of useful features that expose the actions happening
    inside a taskflow engine, which can be useful for developers for debugging,
    for operations folks for monitoring and tracking of the resource actions
    and more...
    """

    def __init__(self, engine,
                 task_listen_for=base.DEFAULT_LISTEN_FOR,
                 flow_listen_for=base.DEFAULT_LISTEN_FOR,
                 retry_listen_for=base.DEFAULT_LISTEN_FOR,
                 logger=LOG):
        super(DynamicLogListener, self).__init__(
            engine,
            task_listen_for=task_listen_for,
            flow_listen_for=flow_listen_for,
            retry_listen_for=retry_listen_for,
            log=logger, fail_formatter=SpecialFormatter(engine))


def get_recovery_flow(task_list, **kwargs):
    """This is used create extension object from provided task_list.

    This method returns the extension object of the each task provided
    in a list using stevedore extension manager.
    """
    extensions = named.NamedExtensionManager(
        'masakari.task_flow.tasks', names=task_list,
        name_order=True, invoke_on_load=True, invoke_kwds=kwargs)
    for extension in extensions.extensions:
        yield extension.obj


def load_taskflow_into_engine(action, nested_flow,
                              process_what):
    book = None
    backend = None
    if PERSISTENCE_BACKEND:
        backend = backends.fetch(PERSISTENCE_BACKEND)
        with contextlib.closing(backend.get_connection()) as conn:
            try:
                book = conn.get_logbook(process_what['notification_uuid'])
            except exceptions.NotFound:
                pass
            if book is None:
                book = models.LogBook(action,
                                      process_what['notification_uuid'])

    return taskflow.engines.load(nested_flow, store=process_what,
                                 backend=backend, book=book)



# Copyright (c) 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from six.moves import http_client
import webob.exc

from masakari.api.openstack import extensions
from masakari.api.openstack import wsgi
from masakari import exception
from masakari.policies import base as base_policies
from masakari.policies import extension_info as extension_policies

ALIAS = 'extensions'
LOG = logging.getLogger(__name__)


class FakeExtension(object):
    def __init__(self, name, alias, description=""):
        self.name = name
        self.alias = alias
        self.__doc__ = description
        self.version = -1


class ExtensionInfoController(wsgi.Controller):

    def __init__(self, extension_info):
        self.extension_info = extension_info

    def _translate(self, ext):
        ext_data = {"name": ext.name,
                    "alias": ext.alias,
                    "description": ext.__doc__,
                    "namespace": "", "updated": "",
                    "links": []}
        return ext_data

    def _create_fake_ext(self, name, alias, description=""):
        return FakeExtension(name, alias, description)

    def _get_extensions(self, context):
        """Filter extensions list based on policy."""

        discoverable_extensions = dict()
        for alias, ext in self.extension_info.get_extensions().items():
            action = ':'.join([
                base_policies.MASAKARI_API, alias, 'discoverable'])
            if context.can(action, fatal=False):
                discoverable_extensions[alias] = ext
            else:
                LOG.debug("Filter out extension %s from discover list",
                          alias)

        return discoverable_extensions

    @extensions.expected_errors(())
    def index(self, req):
        context = req.environ['masakari.context']
        context.can(extension_policies.EXTENSIONS % 'index')
        discoverable_extensions = self._get_extensions(context)
        sorted_ext_list = sorted(discoverable_extensions.items())

        extensions = []
        for _alias, ext in sorted_ext_list:
            extensions.append(self._translate(ext))

        return dict(extensions=extensions)

    @extensions.expected_errors(http_client.NOT_FOUND)
    def show(self, req, id):
        context = req.environ['masakari.context']
        context.can(extension_policies.EXTENSIONS % 'detail')
        try:
            ext = self._get_extensions(context)[id]
        except KeyError:
            raise webob.exc.HTTPNotFound()

        return dict(extension=self._translate(ext))


class ExtensionInfo(extensions.V1APIExtensionBase):
    """Extension information."""

    name = "Extensions"
    alias = ALIAS
    version = 1

    def get_resources(self):
        resources = [
            extensions.ResourceExtension(
                ALIAS, ExtensionInfoController(self.extension_info),
                member_name='extension')]
        return resources

    def get_controller_extensions(self):
        return []


class LoadedExtensionInfo(object):
    """Keep track of all loaded API extensions."""

    def __init__(self):
        self.extensions = {}

    def register_extension(self, ext):
        if not self._check_extension(ext):
            return False

        alias = ext.alias

        if alias in self.extensions:
            raise exception.MasakariException(
                "Found duplicate extension: %s" % alias)
        self.extensions[alias] = ext
        return True

    def _check_extension(self, extension):
        """Checks for required methods in extension objects."""
        try:
            extension.is_valid()
        except AttributeError:
            LOG.exception("Exception loading extension")
            return False

        return True

    def get_extensions(self):
        return self.extensions



# Copyright 2016 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_log import log as logging
from oslo_utils import uuidutils

from masakari.api import utils as api_utils
from masakari import db
from masakari import exception
from masakari import objects
from masakari.objects import base
from masakari.objects import fields

LOG = logging.getLogger(__name__)


@base.MasakariObjectRegistry.register
class Host(base.MasakariPersistentObject, base.MasakariObject,
           base.MasakariObjectDictCompat):

    # Version 1.0: Initial version
    # Version 1.1: Added 'segment_uuid' parameter to 'get_by_uuid' method
    VERSION = '1.1'

    fields = {
        'id': fields.IntegerField(),
        'uuid': fields.UUIDField(),
        'name': fields.StringField(),
        'failover_segment_id': fields.UUIDField(),
        'failover_segment': fields.ObjectField('FailoverSegment'),
        'type': fields.StringField(),
        'reserved': fields.BooleanField(),
        'control_attributes': fields.StringField(),
        'on_maintenance': fields.BooleanField(),
        }

    @staticmethod
    def _from_db_object(context, host, db_host):

        for key in host.fields:
            db_value = db_host.get(key)
            if key == "failover_segment":
                db_value = objects.FailoverSegment._from_db_object(
                    host._context, objects.FailoverSegment(), db_value)

            setattr(host, key, db_value)

        host.obj_reset_changes()
        host._context = context
        return host

    @base.remotable_classmethod
    def get_by_id(cls, context, id):
        db_inst = db.host_get_by_id(context, id)
        return cls._from_db_object(context, cls(), db_inst)

    @base.remotable_classmethod
    def get_by_uuid(cls, context, uuid, segment_uuid=None):
        db_inst = db.host_get_by_uuid(context, uuid, segment_uuid=segment_uuid)
        return cls._from_db_object(context, cls(), db_inst)

    @base.remotable_classmethod
    def get_by_name(cls, context, name):
        db_inst = db.host_get_by_name(context, name)
        return cls._from_db_object(context, cls(), db_inst)

    @base.remotable
    def create(self):
        if self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='create',
                                              reason='already created')
        updates = self.masakari_obj_get_changes()

        if 'uuid' not in updates:
            updates['uuid'] = uuidutils.generate_uuid()
            LOG.debug('Generated uuid %(uuid)s for host',
                      dict(uuid=updates['uuid']))

        if 'failover_segment' in updates:
            raise exception.ObjectActionError(action='create',
                                              reason='failover segment '
                                                     'assigned')

        api_utils.notify_about_host_api(self._context, self,
            action=fields.EventNotificationAction.HOST_CREATE,
            phase=fields.EventNotificationPhase.START)

        db_host = db.host_create(self._context, updates)

        api_utils.notify_about_host_api(self._context, self,
            action=fields.EventNotificationAction.HOST_CREATE,
            phase=fields.EventNotificationPhase.END)

        self._from_db_object(self._context, self, db_host)

    @base.remotable
    def save(self):
        updates = self.masakari_obj_get_changes()
        if 'failover_segment' in updates:
            raise exception.ObjectActionError(action='save',
                                              reason='failover segment '
                                                     'changed')
        updates.pop('id', None)

        api_utils.notify_about_host_api(self._context, self,
            action=fields.EventNotificationAction.HOST_UPDATE,
            phase=fields.EventNotificationPhase.START)

        db_host = db.host_update(self._context, self.uuid, updates)

        api_utils.notify_about_host_api(self._context, self,
            action=fields.EventNotificationAction.HOST_UPDATE,
            phase=fields.EventNotificationPhase.END)

        self._from_db_object(self._context, self, db_host)

    @base.remotable
    def destroy(self):
        if not self.obj_attr_is_set('id'):
            raise exception.ObjectActionError(action='destroy',
                                              reason='already destroyed')
        if not self.obj_attr_is_set('uuid'):
            raise exception.ObjectActionError(action='destroy',
                                              reason='no uuid')

        api_utils.notify_about_host_api(self._context, self,
            action=fields.EventNotificationAction.HOST_DELETE,
            phase=fields.EventNotificationPhase.START)

        db.host_delete(self._context, self.uuid)

        api_utils.notify_about_host_api(self._context, self,
            action=fields.EventNotificationAction.HOST_DELETE,
            phase=fields.EventNotificationPhase.END)

        delattr(self, base.get_attrname('id'))


@base.MasakariObjectRegistry.register
class HostList(base.ObjectListBase, base.MasakariObject):

    VERSION = '1.0'

    fields = {
        'objects': fields.ListOfObjectsField('Host'),
        }

    @base.remotable_classmethod
    def get_all(cls, context, filters=None, sort_keys=None, sort_dirs=None,
                limit=None, marker=None):

        groups = db.host_get_all_by_filters(context, filters=filters,
                                            sort_keys=sort_keys,
                                            sort_dirs=sort_dirs,
                                            limit=limit, marker=marker)

        return base.obj_make_list(context, cls(context), objects.Host, groups)



# Copyright 2017 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import mock
import sys

from masakari.cmd import manage
from masakari import context
from masakari.db import api as db_api
from masakari import test


class DBCommandsTestCase(test.TestCase):

    def setUp(self):
        super(DBCommandsTestCase, self).setUp()
        self.commands = manage.DbCommands()
        self.context = context.get_admin_context()
        sys.argv = ['masakari-manage']

    @mock.patch.object(db_api, 'purge_deleted_rows')
    @mock.patch.object(context, 'get_admin_context')
    def test_purge_command(self, mock_context, mock_db_purge):
        mock_context.return_value = self.context
        self.commands.purge(0, 100)
        mock_db_purge.assert_called_once_with(self.context, 0, 100)

    def test_purge_negative_age_in_days(self):
        ex = self.assertRaises(SystemExit, self.commands.purge, -1, 100)
        self.assertEqual("Must supply a non-negative value for age.", ex.code)

    def test_purge_invalid_age_in_days(self):
        ex = self.assertRaises(SystemExit, self.commands.purge, "test", 100)
        self.assertEqual("Invalid value for age, test", ex.code)

    def test_purge_command_exceeded_age_in_days(self):
        ex = self.assertRaises(SystemExit, self.commands.purge, 1000000, 50)
        self.assertEqual("Maximal age is count of days since epoch.", ex.code)

    def test_purge_invalid_max_rows(self):
        ex = self.assertRaises(SystemExit, self.commands.purge, 0, 0)
        self.assertEqual("Must supply value greater than 0 for max_rows.",
                         ex.code)

    def test_purge_negative_max_rows(self):
        ex = self.assertRaises(SystemExit, self.commands.purge, 0, -5)
        self.assertEqual("Invalid input received: max_rows must be >= -1",
                         ex.code)

    @mock.patch.object(db_api, 'purge_deleted_rows')
    @mock.patch.object(context, 'get_admin_context')
    def test_purge_max_rows(self, mock_context, mock_db_purge):
        mock_context.return_value = self.context
        value = (2 ** 31) - 1
        self.commands.purge(age_in_days=1, max_rows=value)
        mock_db_purge.assert_called_once_with(self.context, 1, value)

    def test_purge_command_exceeded_maximum_rows(self):
        # value(2 ** 31) is greater than max_rows(2147483647) by 1.
        value = 2 ** 31
        ex = self.assertRaises(SystemExit, self.commands.purge, age_in_days=1,
                               max_rows=value)
        expected = "Invalid input received: max_rows must be <= 2147483647"
        self.assertEqual(expected, ex.code)



# Copyright (c) 2016 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#      http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.
"""Unit tests for the DB API."""
from oslo_utils import timeutils

from masakari import context
from masakari import db
from masakari import exception
from masakari import test
from masakari.tests import uuidsentinel

NOW = timeutils.utcnow().replace(microsecond=0)


class ModelsObjectComparatorMixin(object):
    def _dict_from_object(self, obj, ignored_keys):
        if ignored_keys is None:
            ignored_keys = []

        return {k: v for k, v in obj.items()
                if k not in ignored_keys}

    def _assertEqualObjects(self, obj1, obj2, ignored_keys=None):
        obj1 = self._dict_from_object(obj1, ignored_keys)
        obj2 = self._dict_from_object(obj2, ignored_keys)

        self.assertEqual(len(obj1),
                         len(obj2),
                         "Keys mismatch: %s" %
                         str(set(obj1.keys()) ^ set(obj2.keys())))
        for key, value in obj1.items():
            self.assertEqual(value, obj2[key])

    def _assertEqualListsOfObjects(self, objs1, objs2, ignored_keys=None):
        obj_to_dict = lambda o: self._dict_from_object(o, ignored_keys)
        sort_key = lambda d: [d[k] for k in sorted(d)]
        conv_and_sort = lambda obj: sorted(map(obj_to_dict, obj), key=sort_key)

        self.assertEqual(conv_and_sort(objs1), conv_and_sort(objs2))


class FailoverSegmentsTestCase(test.TestCase, ModelsObjectComparatorMixin):

    def setUp(self):
        super(FailoverSegmentsTestCase, self).setUp()
        self.ctxt = context.get_admin_context()

    def _get_fake_values(self):
        return {
            'uuid': uuidsentinel.fake_uuid,
            'name': 'fake_name',
            'service_type': 'fake_service_type',
            'description': 'fake_description',
            'recovery_method': 'auto'
        }

    def _get_fake_values_list(self):
        return [
            {"id": 1, 'name': 'test_1', 'service_type': 'fake_service_type_1',
             'recovery_method': 'auto', 'uuid': uuidsentinel.uuid_1},
            {"id": 2, 'name': 'test_2', 'service_type': 'fake_service_type_2',
             'recovery_method': 'auto', 'uuid': uuidsentinel.uuid_2},
            {"id": 3, 'name': 'test_3', 'service_type': 'fake_service_type_3',
             'recovery_method': 'reserved_host', 'uuid': uuidsentinel.uuid_3}]

    def _create_failover_segment(self, values):
        return db.failover_segment_create(self.ctxt, values)

    def _test_get_failover_segment(self, method, filter):
        failover_segments = [self._create_failover_segment(p)
                             for p in self._get_fake_values_list()]
        for failover_segment in failover_segments:
            real_failover_segment = method(self.ctxt, failover_segment[filter])
            self._assertEqualObjects(failover_segment, real_failover_segment)

    def test_failover_segment_create(self):
        failover_segment = self._create_failover_segment(self.
                                                         _get_fake_values())
        self.assertIsNotNone(failover_segment['id'])
        ignored_keys = ['deleted', 'created_at', 'updated_at', 'deleted_at',
                        'id']
        self._assertEqualObjects(failover_segment, self._get_fake_values(),
                                 ignored_keys)

    def test_failover_segment_get_by_id(self):
        self._test_get_failover_segment(db.failover_segment_get_by_id, 'id')

    def test_failover_segment_get_by_uuid(self):
        self._test_get_failover_segment(
            db.failover_segment_get_by_uuid, 'uuid')

    def test_failover_segment_get_by_name(self):
        self._test_get_failover_segment(
            db.failover_segment_get_by_name, 'name')

    def test_failover_segment_update(self):
        update = {'name': 'updated_name', 'description': 'updated_desc'}
        updated = {'uuid': uuidsentinel.fake_uuid,
                   'name': 'updated_name',
                   'service_type': 'fake_service_type',
                   'description': 'updated_desc',
                   'recovery_method': 'auto'}
        ignored_keys = ['deleted', 'created_at', 'updated_at', 'deleted_at',
                        'id']
        self._create_failover_segment(self._get_fake_values())
        db.failover_segment_update(self.ctxt, uuidsentinel.fake_uuid, update)
        failover_seg_updated = db.failover_segment_get_by_uuid(
            self.ctxt, uuidsentinel.fake_uuid)
        self._assertEqualObjects(updated, failover_seg_updated, ignored_keys)

    def test_failover_segment_delete(self):
        ctxt = context.get_admin_context()
        result = self._create_failover_segment(self._get_fake_values())
        db.failover_segment_delete(ctxt, result['uuid'])
        self.assertRaises(exception.FailoverSegmentNotFound,
                          db.failover_segment_get_by_uuid, self.ctxt,
                          uuidsentinel.fake_uuid)

    def test_failover_segment_get_all_by_filters(self):
        failover_segments = [self._create_failover_segment(p)
                             for p in self._get_fake_values_list()]
        ignored_keys = ['deleted', 'created_at', 'updated_at', 'deleted_at',
                        'id']
        real_failover_segment = db.failover_segment_get_all_by_filters(
            context=self.ctxt,
            filters={'recovery_method': 'auto'},
            marker=1,
            limit=1,
            sort_keys=['id'],
            sort_dirs=['asc'])
        self._assertEqualListsOfObjects([failover_segments[1]],
                                        real_failover_segment, ignored_keys)

    def test_failover_segment_not_found(self):
        self._create_failover_segment(self._get_fake_values())
        self.assertRaises(exception.FailoverSegmentNotFound,
                          db.failover_segment_get_by_id, self.ctxt,
                          500)
        self.assertRaises(exception.FailoverSegmentNotFoundByName,
                          db.failover_segment_get_by_name, self.ctxt,
                          'test')
        self.assertRaises(exception.FailoverSegmentNotFound,
                          db.failover_segment_delete, self.ctxt,
                          uuidsentinel.uuid_4)

    def test_invalid_marker(self):
        [self._create_failover_segment(p)
         for p in self._get_fake_values_list()]
        self.assertRaises(exception.MarkerNotFound,
                          db.failover_segment_get_all_by_filters,
                          context=self.ctxt, marker=6)

    def test_invalid_sort_key(self):
        [self._create_failover_segment(p)
         for p in self._get_fake_values_list()]
        self.assertRaises(exception.InvalidSortKey,
                          db.failover_segment_get_all_by_filters,
                          context=self.ctxt, sort_keys=['invalid_sort_key'])

    def test_create_existing_failover_segment(self):
        self._create_failover_segment(self._get_fake_values())
        self.assertRaises(exception.FailoverSegmentExists,
                          db.failover_segment_create, self.ctxt,
                          self._get_fake_values())

    def test_update_existing_failover_segment(self):
        [self._create_failover_segment(p)
         for p in self._get_fake_values_list()]
        self.assertRaises(exception.FailoverSegmentExists,
                          db.failover_segment_update, self.ctxt,
                          uuidsentinel.uuid_2,
                          {'name': 'test_1'})


class HostsTestCase(test.TestCase, ModelsObjectComparatorMixin):

    def setUp(self):
        super(HostsTestCase, self).setUp()
        self.ctxt = context.get_admin_context()
        self.failover_segment = self._get_failover_segment()

    def _get_failover_segment(self):
        values = {'uuid': uuidsentinel.failover_segment_id,
                  'name': 'fake_segment_name',
                  'service_type': 'fake_service_type',
                  'description': 'fake_description',
                  'recovery_method': 'auto'}
        return db.failover_segment_create(self.ctxt, values)

    def _get_fake_values(self):
        return {
            'uuid': uuidsentinel.fake_uuid,
            'name': 'fake_name',
            'reserved': True,
            'type': 'fake_type',
            'control_attributes': 'fake_control_attr',
            'failover_segment': self.failover_segment,
            'failover_segment_id': uuidsentinel.failover_segment_id,
            'on_maintenance': True
        }

    def _get_fake_values_list(self):
        return [
            {'id': 1, 'uuid': uuidsentinel.uuid_1, 'name': 'name_1',
             'type': 'type_1', 'failover_segment': self.failover_segment,
             'reserved': True, 'control_attributes': 'fake_ctrl_attr_1',
             'failover_segment_id': uuidsentinel.failover_segment_id,
             'on_maintenance': True},
            {'id': 2, 'uuid': uuidsentinel.uuid_2, 'name': 'name_2',
             'type': 'type_1', 'failover_segment': self.failover_segment,
             'reserved': True, 'control_attributes': 'fake_ctrl_attr_2',
             'failover_segment_id': uuidsentinel.failover_segment_id,
             'on_maintenance': True},
            {'id': 3, 'uuid': uuidsentinel.uuid_3, 'name': 'name_3',
             'type': 'type_2', 'failover_segment': self.failover_segment,
             'reserved': True, 'control_attributes': 'fake_ctrl_attr_3',
             'failover_segment_id': uuidsentinel.failover_segment_id,
             'on_maintenance': True}]

    def _create_host(self, values):
        return db.host_create(self.ctxt, values)

    def _test_get_host(self, method, host_uuid_filter,
                       failover_segment_id_filter=None):
        hosts = [self._create_host(p) for p in self._get_fake_values_list()]
        ignored_key = ['failover_segment']
        for host in hosts:
            if failover_segment_id_filter:
                real_host = method(self.ctxt, host[host_uuid_filter],
                                   host[failover_segment_id_filter])
            else:
                real_host = method(self.ctxt, host[host_uuid_filter])
            self._assertEqualObjects(host, real_host, ignored_key)
            self.assertEqual(host['failover_segment'].items(),
                             real_host['failover_segment'].items())

    def test_host_create(self):
        host = self._create_host(self._get_fake_values())
        self.assertIsNotNone(host['id'])
        ignored_keys = ['deleted', 'created_at', 'updated_at', 'deleted_at',
                        'id']
        self._assertEqualObjects(host, self._get_fake_values(), ignored_keys)

    def test_host_get_by_id(self):
        self._test_get_host(db.host_get_by_id, 'id')

    def test_host_get_by_uuid(self):
        self._test_get_host(db.host_get_by_uuid, 'uuid')

    def test_host_get_by_name(self):
        self._test_get_host(db.host_get_by_name, 'name')

    def test_host_get_by_host_uuid_and_failover_segment_id(self):
        self._test_get_host(db.host_get_by_uuid, 'uuid', 'failover_segment_id')

    def test_host_get_by_uuid_filter_by_invalid_failover_segment(self):
        # create hosts under failover_segment
        # 'uuidsentinel.failover_segment_id'
        for host in self._get_fake_values_list():
            self._create_host(host)

        # create one more failover_segment
        values = {'uuid': uuidsentinel.failover_segment_id_1,
                  'name': 'fake_segment_name_1',
                  'service_type': 'fake_service_type',
                  'description': 'fake_description',
                  'recovery_method': 'auto'}
        db.failover_segment_create(self.ctxt, values)

        # try to get host with failover_segment
        # 'uuidsentinel.failover_segment_id_1'
        self.assertRaises(
            exception.HostNotFoundUnderFailoverSegment, db.host_get_by_uuid,
            self.ctxt, uuidsentinel.uuid_1, uuidsentinel.failover_segment_id_1)

    def test_host_update(self):
        update = {'name': 'updated_name', 'type': 'updated_type'}
        updated = {'uuid': uuidsentinel.fake_uuid,
                   'name': 'updated_name',
                   'reserved': True,
                   'type': 'updated_type',
                   'control_attributes': 'fake_control_attr',
                   'failover_segment': self.failover_segment,
                   'failover_segment_id': uuidsentinel.failover_segment_id,
                   'on_maintenance': True}
        ignored_keys = ['deleted', 'created_at', 'updated_at', 'deleted_at',
                        'id', 'failover_segment']
        self._create_host(self._get_fake_values())
        db.host_update(self.ctxt, uuidsentinel.fake_uuid, update)
        host_updated = db.host_get_by_uuid(self.ctxt, uuidsentinel.fake_uuid)
        self._assertEqualObjects(updated, host_updated, ignored_keys)
        self.assertEqual(updated['failover_segment'].items(),
                         host_updated['failover_segment'].items())

    def test_host_delete(self):
        ctxt = context.get_admin_context()
        result = self._create_host(self._get_fake_values())
        db.host_delete(ctxt, result['uuid'])
        self.assertRaises(exception.HostNotFound,
                          db.host_get_by_uuid, self.ctxt,
                          uuidsentinel.fake_uuid)

    def test_host_get_all_by_filters(self):
        hosts = [self._create_host(p) for p in self._get_fake_values_list()]
        ignored_keys = ['deleted', 'created_at', 'updated_at', 'deleted_at',
                        'id', 'failover_segment']
        real_host = db.host_get_all_by_filters(
            context=self.ctxt,
            filters={'type': 'type_1'},
            marker=1,
            limit=1,
            sort_keys=['id'],
            sort_dirs=['asc'])
        self._assertEqualListsOfObjects([hosts[1]], real_host, ignored_keys)

    def test_host_get_all_by_filter_on_maintenance(self):
        for p in self._get_fake_values_list():
            # create temporary reserved_hosts, all are on maintenance
            self._create_host(p)

        # create one more reserved_host which is not on maintenance
        temp_host = self._get_fake_values()
        temp_host['on_maintenance'] = False
        self._create_host(temp_host)

        ignored_keys = ['deleted', 'created_at', 'updated_at', 'deleted_at',
                        'id', 'failover_segment']
        real_host = db.host_get_all_by_filters(
            context=self.ctxt,
            filters={'on_maintenance': False},
            marker=1,
            limit=1,
            sort_keys=['id'],
            sort_dirs=['asc'])
        self._assertEqualListsOfObjects([temp_host], real_host, ignored_keys)

    def test_host_not_found(self):
        self._create_host(self._get_fake_values())
        self.assertRaises(exception.HostNotFound,
                          db.host_get_by_id, self.ctxt,
                          5)
        self.assertRaises(exception.HostNotFoundByName,
                          db.host_get_by_name, self.ctxt,
                          'test')
        self.assertRaises(exception.HostNotFound,
                          db.host_delete, self.ctxt,
                          uuidsentinel.uuid_4)

    def test_invalid_marker(self):
        [self._create_host(p) for p in self._get_fake_values_list()]
        self.assertRaises(exception.MarkerNotFound,
                          db.host_get_all_by_filters, context=self.ctxt,
                          marker=6)

    def test_invalid_sort_key(self):
        [self._create_host(p) for p in self._get_fake_values_list()]
        self.assertRaises(exception.InvalidSortKey,
                          db.host_get_all_by_filters, context=self.ctxt,
                          sort_keys=['invalid_sort_key'])

    def test_create_existing_host(self):
        self._create_host(self._get_fake_values())
        self.assertRaises(exception.HostExists,
                          db.host_create, self.ctxt,
                          self._get_fake_values())

    def test_update_existing_host(self):
        [self._create_host(p) for p in self._get_fake_values_list()]
        self.assertRaises(exception.HostExists, db.host_update,
                          self.ctxt, uuidsentinel.uuid_2, {'name': 'name_1'})


class NotificationsTestCase(test.TestCase, ModelsObjectComparatorMixin):

    def setUp(self):
        super(NotificationsTestCase, self).setUp()
        self.ctxt = context.get_admin_context()

    def _get_fake_values(self):
        return {
            'notification_uuid': uuidsentinel.notification,
            'generated_time': NOW,
            'source_host_uuid': uuidsentinel.source_host,
            'type': 'fake_type',
            'payload': 'fake_payload',
            'status': 'new'
        }

    def _get_fake_values_list(self):
        return [
            {'id': 1, 'notification_uuid': uuidsentinel.notification_1,
             'generated_time': NOW,
             'source_host_uuid': uuidsentinel.s_host_1, 'type': 'fake_type',
             'payload': 'fake_payload', 'status': 'new'},
            {'id': 2, 'notification_uuid': uuidsentinel.notification_2,
             'generated_time': NOW,
             'source_host_uuid': uuidsentinel.s_host_2, 'type': 'fake_type',
             'payload': 'fake_payload', 'status': 'new'},
            {'id': 3, 'notification_uuid': uuidsentinel.notification_3,
             'generated_time': NOW,
             'source_host_uuid': uuidsentinel.s_host_3, 'type': 'fake_type',
             'payload': 'fake_payload', 'status': 'failed'}]

    def _create_notification(self, values):
        return db.notification_create(self.ctxt, values)

    def _test_get_notification(self, method, filter):
        notifications = [self._create_notification(p)
                         for p in self._get_fake_values_list()]
        for notification in notifications:
            real_notification = method(self.ctxt, notification[filter])
            self._assertEqualObjects(notification, real_notification)

    def test_notification_create(self):
        notification = self._create_notification(self._get_fake_values())
        self.assertIsNotNone(notification['id'])
        ignored_keys = ['deleted', 'created_at', 'updated_at', 'deleted_at',
                        'id']
        self._assertEqualObjects(notification, self._get_fake_values(),
                                 ignored_keys)

    def test_notification_get_by_id(self):
        self._test_get_notification(db.notification_get_by_id, 'id')

    def test_notification_get_by_uuid(self):
        self._test_get_notification(db.notification_get_by_uuid,
                                    'notification_uuid')

    def test_notification_update(self):
        update = {'type': 'updated_type', 'payload': 'updated_payload'}
        updated = {'notification_uuid': uuidsentinel.notification,
                   'generated_time': NOW,
                   'source_host_uuid': uuidsentinel.source_host,
                   'type': 'updated_type',
                   'payload': 'updated_payload',
                   'status': 'new'}
        ignored_keys = ['deleted', 'created_at', 'updated_at', 'deleted_at',
                        'id']
        self._create_notification(self._get_fake_values())
        db.notification_update(self.ctxt, uuidsentinel.notification, update)
        notification_updated = db.notification_get_by_uuid(
            self.ctxt, uuidsentinel.notification)
        self._assertEqualObjects(updated, notification_updated, ignored_keys)

    def test_notification_delete(self):
        ctxt = context.get_admin_context()
        result = self._create_notification(self._get_fake_values())
        db.notification_delete(ctxt, result['notification_uuid'])
        self.assertRaises(exception.NotificationNotFound,
                          db.notification_get_by_uuid, self.ctxt,
                          uuidsentinel.notification)

    def test_notification_get_all_by_filters(self):
        notifications = [self._create_notification(p)
                         for p in self._get_fake_values_list()]
        ignored_keys = ['deleted', 'created_at', 'updated_at', 'deleted_at',
                        'id']
        real_notification = db.notifications_get_all_by_filters(
            context=self.ctxt,
            filters={'status': 'new'},
            marker=1,
            limit=1,
            sort_keys=['id'],
            sort_dirs=['asc'])
        self._assertEqualListsOfObjects([notifications[1]],
                                        real_notification, ignored_keys)

    def test_notification_not_found(self):
        self._create_notification(self._get_fake_values())
        self.assertRaises(exception.NotificationNotFound,
                          db.notification_get_by_id, self.ctxt,
                          500)
        self.assertRaises(exception.NotificationNotFound,
                          db.notification_delete, self.ctxt,
                          uuidsentinel.fake_uuid)

    def test_invalid_marker(self):
        [self._create_notification(p) for p in self._get_fake_values_list()]
        self.assertRaises(exception.MarkerNotFound,
                          db.notifications_get_all_by_filters,
                          context=self.ctxt, marker=6)

    def test_invalid_sort_key(self):
        [self._create_notification(p) for p in self._get_fake_values_list()]
        self.assertRaises(exception.InvalidSortKey,
                          db.notifications_get_all_by_filters,
                          context=self.ctxt, sort_keys=['invalid_sort_key'])



#    Copyright 2016 NTT DATA
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import copy

import mock
from oslo_utils import timeutils

from masakari.api import utils as api_utils
from masakari import exception
from masakari.objects import fields
from masakari.objects import segment
from masakari.tests.unit.objects import test_objects
from masakari.tests import uuidsentinel


NOW = timeutils.utcnow().replace(microsecond=0)

fake_segment = {
    'created_at': NOW,
    'updated_at': None,
    'deleted_at': None,
    'deleted': False,
    'id': 123,
    'uuid': uuidsentinel.fake_segment,
    'name': 'foo-segment',
    'service_type': 'COMPUTE',
    'description': 'fake-description',
    'recovery_method': 'auto'
    }


class TestFailoverSegmentObject(test_objects._LocalTest):

    @mock.patch('masakari.db.failover_segment_get_by_name')
    def test_get_by_name(self, mock_api_get):

        mock_api_get.return_value = fake_segment

        segment_obj = segment.FailoverSegment.get_by_name(self.context,
                                                          'foo-segment')
        self.compare_obj(segment_obj, fake_segment)

        mock_api_get.assert_called_once_with(self.context, 'foo-segment')

    @mock.patch('masakari.db.failover_segment_get_by_uuid')
    def test_get_by_uuid(self, mock_api_get):

        mock_api_get.return_value = fake_segment

        segment_obj = (segment.FailoverSegment.
                       get_by_uuid(self.context, uuidsentinel.fake_segment))
        self.compare_obj(segment_obj, fake_segment)

        mock_api_get.assert_called_once_with(self.context,
                                             uuidsentinel.fake_segment)

    @mock.patch('masakari.db.failover_segment_get_by_id')
    def test_get_by_id(self, mock_api_get):

        mock_api_get.return_value = fake_segment
        fake_id = 123
        segment_obj = segment.FailoverSegment.get_by_id(self.context, fake_id)
        self.compare_obj(segment_obj, fake_segment)

        mock_api_get.assert_called_once_with(self.context, fake_id)

    def _segment_create_attribute(self):

        segment_obj = segment.FailoverSegment(context=self.context)
        segment_obj.name = 'foo-segment'
        segment_obj.description = 'keydata'
        segment_obj.service_type = 'fake-user'
        segment_obj.recovery_method = 'auto'
        segment_obj.uuid = uuidsentinel.fake_segment

        return segment_obj

    @mock.patch.object(api_utils, 'notify_about_segment_api')
    @mock.patch('masakari.db.failover_segment_create')
    def test_create(self, mock_segment_create, mock_notify_about_segment_api):
        mock_segment_create.return_value = fake_segment

        segment_obj = self._segment_create_attribute()
        segment_obj.create()
        self.compare_obj(segment_obj, fake_segment)

        mock_segment_create.assert_called_once_with(self.context, {
            'uuid': uuidsentinel.fake_segment, 'name': 'foo-segment',
            'description': 'keydata', 'service_type': 'fake-user',
            'recovery_method': 'auto'})
        action = fields.EventNotificationAction.SEGMENT_CREATE
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, segment_obj, action=action,
                      phase=phase_start),
            mock.call(self.context, segment_obj, action=action,
                      phase=phase_end)]
        mock_notify_about_segment_api.assert_has_calls(notify_calls)

    @mock.patch.object(api_utils, 'notify_about_segment_api')
    @mock.patch('masakari.db.failover_segment_create')
    def test_recreate_fails(self, mock_segment_create,
                            mock_notify_about_segment_api):
        mock_segment_create.return_value = fake_segment

        segment_obj = self._segment_create_attribute()
        segment_obj.create()
        self.assertRaises(exception.ObjectActionError, segment_obj.create)

        mock_segment_create.assert_called_once_with(self.context, {
            'uuid': uuidsentinel.fake_segment, 'name': 'foo-segment',
            'description': 'keydata', 'service_type': 'fake-user',
            'recovery_method': 'auto'})
        action = fields.EventNotificationAction.SEGMENT_CREATE
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, segment_obj, action=action,
                      phase=phase_start),
            mock.call(self.context, segment_obj, action=action,
                      phase=phase_end)]
        mock_notify_about_segment_api.assert_has_calls(notify_calls)

    @mock.patch.object(api_utils, 'notify_about_segment_api')
    @mock.patch('masakari.db.failover_segment_delete')
    def test_destroy(self, mock_segment_destroy,
                     mock_notify_about_segment_api):
        segment_obj = self._segment_create_attribute()
        segment_obj.id = 123
        segment_obj.destroy()

        mock_segment_destroy.assert_called_once_with(
            self.context, uuidsentinel.fake_segment)
        action = fields.EventNotificationAction.SEGMENT_DELETE
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, segment_obj, action=action,
                      phase=phase_start),
            mock.call(self.context, segment_obj, action=action,
                      phase=phase_end)]
        mock_notify_about_segment_api.assert_has_calls(notify_calls)

    @mock.patch.object(api_utils, 'notify_about_segment_api')
    @mock.patch('masakari.db.failover_segment_delete')
    def test_destroy_failover_segment_found(self, mock_segment_destroy,
                                            mock_notify_about_segment_api):
        mock_segment_destroy.side_effect = exception.FailoverSegmentNotFound(
            id=123)
        segment_obj = self._segment_create_attribute()
        segment_obj.id = 123
        self.assertRaises(exception.FailoverSegmentNotFound,
                          segment_obj.destroy)
        action = fields.EventNotificationAction.SEGMENT_DELETE
        phase_start = fields.EventNotificationPhase.START
        notify_calls = [
            mock.call(self.context, segment_obj, action=action,
                      phase=phase_start)]
        mock_notify_about_segment_api.assert_has_calls(notify_calls)

    @mock.patch('masakari.db.failover_segment_get_all_by_filters')
    def test_get_segment_by_recovery_method(self, mock_api_get):
        fake_segment2 = copy.deepcopy(fake_segment)
        fake_segment2['name'] = 'fake_segment2'

        mock_api_get.return_value = [fake_segment2, fake_segment]

        segment_result = (segment.FailoverSegmentList.
                          get_all(self.context,
                                  filters={'recovery_method': 'auto'}))
        self.assertEqual(2, len(segment_result))
        self.compare_obj(segment_result[0], fake_segment2)
        self.compare_obj(segment_result[1], fake_segment)
        mock_api_get.assert_called_once_with(self.context, filters={
            'recovery_method': 'auto'
        }, limit=None, marker=None, sort_dirs=None, sort_keys=None)

    @mock.patch('masakari.db.failover_segment_get_all_by_filters')
    def test_get_segment_by_service_type(self, mock_api_get):
        fake_segment2 = copy.deepcopy(fake_segment)
        fake_segment2['name'] = 'fake_segment'

        mock_api_get.return_value = [fake_segment2, fake_segment]

        segment_result = (segment.FailoverSegmentList.
                          get_all(self.context,
                                  filters={'service_type': 'COMPUTE'}))
        self.assertEqual(2, len(segment_result))
        self.compare_obj(segment_result[0], fake_segment2)
        self.compare_obj(segment_result[1], fake_segment)
        mock_api_get.assert_called_once_with(self.context, filters={
            'service_type': 'COMPUTE'
        }, limit=None, marker=None, sort_dirs=None, sort_keys=None)

    @mock.patch('masakari.db.failover_segment_get_all_by_filters')
    def test_get_limit_and_marker_invalid_marker(self, mock_api_get):
        segment_name = 'unknown_segment'
        mock_api_get.side_effect = exception.MarkerNotFound(marker=segment_name
                                                            )

        self.assertRaises(exception.MarkerNotFound,
                          segment.FailoverSegmentList.get_all,
                          self.context, limit=5, marker=segment_name)

    @mock.patch.object(api_utils, 'notify_about_segment_api')
    @mock.patch('masakari.db.failover_segment_update')
    def test_save(self, mock_segment_update, mock_notify_about_segment_api):

        mock_segment_update.return_value = fake_segment

        segment_object = segment.FailoverSegment(context=self.context)
        segment_object.name = "foo-segment"
        segment_object.id = 123
        segment_object.uuid = uuidsentinel.fake_segment
        segment_object.save()

        self.compare_obj(segment_object, fake_segment)
        self.assertTrue(mock_segment_update.called)
        action = fields.EventNotificationAction.SEGMENT_UPDATE
        phase_start = fields.EventNotificationPhase.START
        phase_end = fields.EventNotificationPhase.END
        notify_calls = [
            mock.call(self.context, segment_object, action=action,
                      phase=phase_start),
            mock.call(self.context, segment_object, action=action,
                      phase=phase_end)]
        mock_notify_about_segment_api.assert_has_calls(notify_calls)

    @mock.patch.object(api_utils, 'notify_about_segment_api')
    @mock.patch('masakari.db.failover_segment_update')
    def test_save_failover_segment_not_found(self, mock_segment_update,
                                             mock_notify_about_segment_api):

        mock_segment_update.side_effect = (
            exception.FailoverSegmentNotFound(id=uuidsentinel.fake_segment))

        segment_object = segment.FailoverSegment(context=self.context)
        segment_object.name = "foo-segment"
        segment_object.id = 123
        segment_object.uuid = uuidsentinel.fake_segment

        self.assertRaises(exception.FailoverSegmentNotFound,
                          segment_object.save)
        action = fields.EventNotificationAction.SEGMENT_UPDATE
        phase_start = fields.EventNotificationPhase.START
        notify_calls = [
            mock.call(self.context, segment_object, action=action,
                      phase=phase_start)]
        mock_notify_about_segment_api.assert_has_calls(notify_calls)

    @mock.patch.object(api_utils, 'notify_about_segment_api')
    @mock.patch('masakari.db.failover_segment_update')
    def test_save_failover_segment_already_exists(self, mock_segment_update,
                                            mock_notify_about_segment_api):

        mock_segment_update.side_effect = (
            exception.FailoverSegmentExists(name="foo-segment"))

        segment_object = segment.FailoverSegment(context=self.context)
        segment_object.name = "foo-segment"
        segment_object.id = 123
        segment_object.uuid = uuidsentinel.fake_segment

        self.assertRaises(exception.FailoverSegmentExists, segment_object.save)
        action = fields.EventNotificationAction.SEGMENT_UPDATE
        phase_start = fields.EventNotificationPhase.START
        notify_calls = [
            mock.call(self.context, segment_object, action=action,
                      phase=phase_start)]
        mock_notify_about_segment_api.assert_has_calls(notify_calls)



#    Copyright 2016 NTT DATA
#    All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import datetime

import iso8601
from oslo_versionedobjects import exception as ovo_exc
import six

from masakari.objects import fields
from masakari import test
from masakari import utils


class FakeFieldType(fields.FieldType):
    def coerce(self, obj, attr, value):
        return '*%s*' % value

    def to_primitive(self, obj, attr, value):
        return '!%s!' % value

    def from_primitive(self, obj, attr, value):
        return value[1:-1]


class FakeEnum(fields.Enum):
    FROG = "frog"
    PLATYPUS = "platypus"
    ALLIGATOR = "alligator"

    ALL = (FROG, PLATYPUS, ALLIGATOR)

    def __init__(self, **kwargs):
        super(FakeEnum, self).__init__(valid_values=FakeEnum.ALL,
                                       **kwargs)


class FakeEnumAlt(fields.Enum):
    FROG = "frog"
    PLATYPUS = "platypus"
    AARDVARK = "aardvark"

    ALL = (FROG, PLATYPUS, AARDVARK)

    def __init__(self, **kwargs):
        super(FakeEnumAlt, self).__init__(valid_values=FakeEnumAlt.ALL,
                                          **kwargs)


class FakeEnumField(fields.BaseEnumField):
    AUTO_TYPE = FakeEnum()


class FakeEnumAltField(fields.BaseEnumField):
    AUTO_TYPE = FakeEnumAlt()


class TestField(test.NoDBTestCase):
    def setUp(self):
        super(TestField, self).setUp()
        self.field = fields.Field(FakeFieldType())
        self.coerce_good_values = [('foo', '*foo*')]
        self.coerce_bad_values = []
        self.to_primitive_values = [('foo', '!foo!')]
        self.from_primitive_values = [('!foo!', 'foo')]

    def test_coerce_good_values(self):
        for in_val, out_val in self.coerce_good_values:
            self.assertEqual(out_val, self.field.coerce('obj', 'attr', in_val))

    def test_coerce_bad_values(self):
        for in_val in self.coerce_bad_values:
            self.assertRaises((TypeError, ValueError),
                              self.field.coerce, 'obj', 'attr', in_val)

    def test_to_primitive(self):
        for in_val, prim_val in self.to_primitive_values:
            self.assertEqual(prim_val, self.field.to_primitive('obj', 'attr',
                                                               in_val))

    def test_from_primitive(self):
        class ObjectLikeThing(object):
            _context = 'context'

        for prim_val, out_val in self.from_primitive_values:
            self.assertEqual(out_val,
                             self.field.from_primitive(ObjectLikeThing,
                                                       'attr', prim_val))

    def test_stringify(self):
        self.assertEqual('123', self.field.stringify(123))


class TestString(TestField):
    def setUp(self):
        super(TestString, self).setUp()
        self.field = fields.StringField()
        self.coerce_good_values = [('foo', 'foo'), (1, '1'), (True, 'True')]
        if six.PY2:
            self.coerce_good_values.append((int(1), '1'))
        self.coerce_bad_values = [None]
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]

    def test_stringify(self):
        self.assertEqual("'123'", self.field.stringify(123))


class TestBaseEnum(TestField):
    def setUp(self):
        super(TestBaseEnum, self).setUp()
        self.field = FakeEnumField()
        self.coerce_good_values = [('frog', 'frog'),
                                   ('platypus', 'platypus'),
                                   ('alligator', 'alligator')]
        self.coerce_bad_values = ['aardvark', 'wookie']
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]

    def test_stringify(self):
        self.assertEqual("'platypus'", self.field.stringify('platypus'))

    def test_stringify_invalid(self):
        self.assertRaises(ValueError, self.field.stringify, 'aardvark')

    def test_fingerprint(self):
        field1 = FakeEnumField()
        field2 = FakeEnumAltField()
        self.assertNotEqual(str(field1), str(field2))


class TestEnum(TestField):
    def setUp(self):
        super(TestEnum, self).setUp()
        self.field = fields.EnumField(
            valid_values=['foo', 'bar', 1, 1, True])
        self.coerce_good_values = [('foo', 'foo'), (1, '1'), (True, 'True')]
        if six.PY2:
            self.coerce_good_values.append((int(1), '1'))
        self.coerce_bad_values = ['boo', 2, False]
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]

    def test_stringify(self):
        self.assertEqual("'foo'", self.field.stringify('foo'))

    def test_stringify_invalid(self):
        self.assertRaises(ValueError, self.field.stringify, '123')

    def test_fingerprint(self):
        field1 = fields.EnumField(valid_values=['foo', 'bar'])
        field2 = fields.EnumField(valid_values=['foo', 'bar1'])
        self.assertNotEqual(str(field1), str(field2))

    def test_without_valid_values(self):
        self.assertRaises(ovo_exc.EnumValidValuesInvalidError,
                          fields.EnumField, 1)

    def test_with_empty_values(self):
        self.assertRaises(ovo_exc.EnumRequiresValidValuesError,
                          fields.EnumField, [])


class TestDictOfStrings(TestField):
    def setUp(self):
        super(TestDictOfStrings, self).setUp()
        self.field = fields.DictOfStringsField()
        self.coerce_good_values = [({'foo': 'bar'}, {'foo': 'bar'}),
                                   ({'foo': 1}, {'foo': '1'})]
        self.coerce_bad_values = [{1: 'bar'}, {'foo': None}, 'foo']
        self.to_primitive_values = [({'foo': 'bar'}, {'foo': 'bar'})]
        self.from_primitive_values = [({'foo': 'bar'}, {'foo': 'bar'})]

    def test_stringify(self):
        self.assertEqual("{key='val'}", self.field.stringify({'key': 'val'}))


class TestInteger(TestField):
    def setUp(self):
        super(TestInteger, self).setUp()
        self.field = fields.IntegerField()
        self.coerce_good_values = [(1, 1), ('1', 1)]
        self.coerce_bad_values = ['foo', None]
        self.to_primitive_values = self.coerce_good_values[0:1]
        self.from_primitive_values = self.coerce_good_values[0:1]


class TestBoolean(TestField):
    def setUp(self):
        super(TestBoolean, self).setUp()
        self.field = fields.BooleanField()
        self.coerce_good_values = [(True, True), (False, False), (1, True),
                                   ('foo', True), (0, False), ('', False)]
        self.coerce_bad_values = []
        self.to_primitive_values = self.coerce_good_values[0:2]
        self.from_primitive_values = self.coerce_good_values[0:2]


class TestDateTime(TestField):
    def setUp(self):
        super(TestDateTime, self).setUp()
        self.dt = datetime.datetime(2016, 11, 5, tzinfo=iso8601.UTC)
        self.field = fields.DateTimeField()
        self.coerce_good_values = [(self.dt, self.dt),
                                   (utils.isotime(self.dt), self.dt)]
        self.coerce_bad_values = [1, 'foo']
        self.to_primitive_values = [(self.dt, utils.isotime(self.dt))]
        self.from_primitive_values = [(utils.isotime(self.dt), self.dt)]

    def test_stringify(self):
        self.assertEqual(
            '2016-11-05T18:00:00Z',
            self.field.stringify(
                datetime.datetime(2016, 11, 5, 18, 0, 0,
                                  tzinfo=iso8601.iso8601.UTC)))



# Copyright 2016 NTT DATA.
# All rights reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from six.moves import http_client as http
from webob import exc

from masakari.api.openstack import common
from masakari.api.openstack import extensions
from masakari.api.openstack.ha.schemas import segments as schema
from masakari.api.openstack import wsgi
from masakari.api import validation
from masakari import exception
from masakari.ha import api as segment_api
from masakari.policies import segments as segment_policies


ALIAS = 'segments'


class SegmentsController(wsgi.Controller):
    """Segments controller for the OpenStack API."""

    def __init__(self):
        self.api = segment_api.FailoverSegmentAPI()

    @extensions.expected_errors((http.BAD_REQUEST, http.FORBIDDEN))
    def index(self, req):
        """Returns a summary list of failover segments."""
        context = req.environ['masakari.context']
        context.can(segment_policies.SEGMENTS % 'index')

        try:
            limit, marker = common.get_limit_and_marker(req)
            sort_keys, sort_dirs = common.get_sort_params(req.params)

            filters = {}
            if 'recovery_method' in req.params:
                filters['recovery_method'] = req.params['recovery_method']
            if 'service_type' in req.params:
                filters['service_type'] = req.params['service_type']

            segments = self.api.get_all(context, filters=filters,
                                        sort_keys=sort_keys,
                                        sort_dirs=sort_dirs, limit=limit,
                                        marker=marker)
        except exception.MarkerNotFound as e:
            raise exc.HTTPBadRequest(explanation=e.format_message())
        except exception.Invalid as e:
            raise exc.HTTPBadRequest(explanation=e.format_message())

        return {'segments': segments}

    @extensions.expected_errors((http.FORBIDDEN, http.NOT_FOUND))
    def show(self, req, id):
        """Return data about the given segment id."""
        context = req.environ['masakari.context']
        context.can(segment_policies.SEGMENTS % 'detail')

        try:
            segment = self.api.get_segment(context, id)
        except exception.FailoverSegmentNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        return {'segment': segment}

    @wsgi.response(http.CREATED)
    @extensions.expected_errors((http.FORBIDDEN, http.CONFLICT))
    @validation.schema(schema.create)
    def create(self, req, body):
        """Creates a new failover segment."""
        context = req.environ['masakari.context']
        context.can(segment_policies.SEGMENTS % 'create')

        segment_data = body['segment']
        try:
            segment = self.api.create_segment(context, segment_data)
        except exception.FailoverSegmentExists as e:
            raise exc.HTTPConflict(explanation=e.format_message())
        return {'segment': segment}

    @extensions.expected_errors((http.FORBIDDEN, http.NOT_FOUND,
                                 http.CONFLICT))
    @validation.schema(schema.update)
    def update(self, req, id, body):
        """Updates the existing segment."""
        context = req.environ['masakari.context']
        context.can(segment_policies.SEGMENTS % 'update')
        segment_data = body['segment']

        try:
            segment = self.api.update_segment(context, id, segment_data)
        except exception.FailoverSegmentNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        except (exception.FailoverSegmentExists, exception.Conflict) as e:
            raise exc.HTTPConflict(explanation=e.format_message())

        return {'segment': segment}

    @wsgi.response(http.NO_CONTENT)
    @extensions.expected_errors((http.FORBIDDEN, http.NOT_FOUND,
                                 http.CONFLICT))
    def delete(self, req, id):
        """Removes a segment by uuid."""
        context = req.environ['masakari.context']
        context.can(segment_policies.SEGMENTS % 'delete')

        try:
            self.api.delete_segment(context, id)
        except exception.FailoverSegmentNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        except exception.Conflict as e:
            raise exc.HTTPConflict(explanation=e.format_message())


class Segments(extensions.V1APIExtensionBase):
    """Segments Extension."""
    name = "Segments"
    alias = ALIAS
    version = 1

    def get_resources(self):
        member_actions = {'action': 'POST'}

        resources = [
            extensions.ResourceExtension(ALIAS,
                                         SegmentsController(),
                                         member_name='segment',
                                         member_actions=member_actions)
            ]
        return resources

    def get_controller_extensions(self):
        return []



# Copyright (c) 2018 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or
# implied.
# See the License for the specific language governing permissions and
# limitations under the License.

from masakari.objects import base
from masakari.objects import fields
from masakari import rpc


@base.MasakariObjectRegistry.register_if(False)
class NotificationObject(base.MasakariObject):
    """Base class for every notification related versioned object."""
    # Version 1.0: Initial version
    VERSION = '1.0'

    def __init__(self, **kwargs):
        super(NotificationObject, self).__init__(**kwargs)
        # The notification objects are created on the fly when masakari emits
        # the notification. This causes that every object shows every field as
        # changed. We don't want to send this meaningless information so we
        # reset the object after creation.
        self.obj_reset_changes(recursive=False)


@base.MasakariObjectRegistry.register_notification
class EventType(NotificationObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'action': fields.EventNotificationActionField(nullable=False),
        'phase': fields.EventNotificationPhaseField(nullable=True),
    }

    def to_notification_event_type_field(self):
        """Serialize the object to the wire format."""
        s = '%s' % (self.action)
        if self.obj_attr_is_set('phase'):
            s += '.%s' % self.phase
        return s


@base.MasakariObjectRegistry.register_if(False)
class NotificationPayloadBase(NotificationObject):
    """Base class for the payload of versioned notifications."""
    # SCHEMA defines how to populate the payload fields. It is a dictionary
    # where every key value pair has the following format:
    # <payload_field_name>: (<data_source_name>,
    #                        <field_of_the_data_source>)
    # The <payload_field_name> is the name where the data will be stored in the
    # payload object, this field has to be defined as a field of the payload.
    # The <data_source_name> shall refer to name of the parameter passed as
    # kwarg to the payload's populate_schema() call and this object will be
    # used as the source of the data. The <field_of_the_data_source> shall be
    # a valid field of the passed argument.
    # The SCHEMA needs to be applied with the populate_schema() call before the
    # notification can be emitted.
    # The value of the payload.<payload_field_name> field will be set by the
    # <data_source_name>.<field_of_the_data_source> field. The
    # <data_source_name> will not be part of the payload object internal or
    # external representation.
    # Payload fields that are not set by the SCHEMA can be filled in the same
    # way as in any versioned object.
    SCHEMA = {}
    # Version 1.0: Initial version
    VERSION = '1.0'

    def __init__(self, **kwargs):
        super(NotificationPayloadBase, self).__init__(**kwargs)
        self.populated = not self.SCHEMA

    def populate_schema(self, **kwargs):
        """Populate the object based on the SCHEMA and the source objects

        :param kwargs: A dict contains the source object at the key defined in
                       the SCHEMA
        """
        for key, (obj, field) in self.SCHEMA.items():
            source = kwargs[obj]
            if source.obj_attr_is_set(field):
                setattr(self, key, getattr(source, field))
        self.populated = True

        # the schema population will create changed fields but we don't need
        # this information in the notification
        self.obj_reset_changes(recursive=False)


@base.MasakariObjectRegistry.register_notification
class NotificationPublisher(NotificationObject):
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'host': fields.StringField(nullable=False),
        'binary': fields.StringField(nullable=False),
    }

    @classmethod
    def from_service_obj(cls, service):
        return cls(host=service.host, binary=service.binary)


@base.MasakariObjectRegistry.register_if(False)
class NotificationBase(NotificationObject):
    """Base class for versioned notifications.

    Every subclass shall define a 'payload' field.
    """
    # Version 1.0: Initial version
    VERSION = '1.0'

    fields = {
        'priority': fields.EventNotificationPriorityField(),
        'event_type': fields.ObjectField('EventType'),
        'publisher': fields.ObjectField('NotificationPublisher'),
    }

    def _emit(self, context, event_type, publisher_id, payload):
        notifier = rpc.get_versioned_notifier(publisher_id)
        notify = getattr(notifier, self.priority)
        notify(context, event_type=event_type, payload=payload)

    def emit(self, context):
        """Send the notification."""
        assert self.payload.populated

        # notification payload will be a newly populated object
        # therefore every field of it will look changed so this does not carry
        # any extra information so we drop this from the payload.
        self.payload.obj_reset_changes(recursive=False)

        self._emit(context,
            event_type=self.event_type.to_notification_event_type_field(),
            publisher_id='%s:%s' %
                         (self.publisher.binary,
                          self.publisher.host),
            payload=self.payload.obj_to_primitive())


def notification_sample(sample):
    """Class decorator for documentation generation purposes.

    This is to attach the notification sample information
    to the notification object for documentation generation purposes.

    :param sample: the path of the sample json file relative to the
                   doc/notification_samples/ directory in the masakari
                   repository root.
    """
    def wrap(cls):
        if not getattr(cls, 'samples', None):
            cls.samples = [sample]
        else:
            cls.samples.append(sample)
        return cls
    return wrap



# Copyright (c) 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from keystoneauth1 import loading as ks_loading
from oslo_config import cfg

nova_opts = [
    cfg.StrOpt('nova_catalog_admin_info',
               default='compute:nova:publicURL',
               help='Match this value when searching for nova in the '
                    'service catalog. Format is: separated values of '
                    'the form: '
                    '<service_type>:<service_name>:<endpoint_type>'),
    cfg.StrOpt('os_region_name',
               help='Region name of this node'),
    cfg.StrOpt('nova_ca_certificates_file',
               help='Location of ca certificates file to use for nova client '
                    'requests.'),
    cfg.BoolOpt('nova_api_insecure',
                default=False,
                help='Allow to perform insecure SSL requests to nova'),
    cfg.StrOpt('os_privileged_user_name',
               help='OpenStack privileged account username. Used for requests '
                    'to other services (such as Nova) that require an account '
                    'with special rights.'),
    cfg.StrOpt('os_privileged_user_password',
               help='Password associated with the OpenStack privileged '
                    'account.',
               secret=True),
    cfg.StrOpt('os_privileged_user_tenant',
               help='Tenant name associated with the OpenStack privileged '
                    'account.'),
    cfg.URIOpt('os_privileged_user_auth_url',
               help='Auth URL associated with the OpenStack privileged '
                    'account.'),
    cfg.StrOpt('os_user_domain_name',
               default="default",
               help='User domain name associated with the OpenStack '
                    'privileged account.'),
    cfg.StrOpt('os_project_domain_name',
               default="default",
               help='Project domain name associated with the OpenStack '
                    'privileged account.'),
]


def register_opts(conf):
    conf.register_opts(nova_opts)
    ks_loading.register_session_conf_options(conf, 'DEFAULT')


def list_opts():
    return {
        'DEFAULT': nova_opts
    }



# Copyright 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import collections
import functools

import oslo_messaging as messaging
from oslo_serialization import jsonutils


NOTIFICATIONS = []
VERSIONED_NOTIFICATIONS = []


def reset():
    del NOTIFICATIONS[:]
    del VERSIONED_NOTIFICATIONS[:]


FakeMessage = collections.namedtuple('Message',
                                     ['publisher_id', 'priority',
                                      'event_type', 'payload', 'context'])


class FakeNotifier(object):

    def __init__(self, transport, publisher_id, serializer=None):
        self.transport = transport
        self.publisher_id = publisher_id
        self._serializer = serializer or messaging.serializer.NoOpSerializer()

        for priority in ['debug', 'info', 'warn', 'error', 'critical']:
            setattr(self, priority,
                    functools.partial(self._notify, priority.upper()))

    def prepare(self, publisher_id=None):
        if publisher_id is None:
            publisher_id = self.publisher_id
        return self.__class__(self.transport, publisher_id,
                              serializer=self._serializer)

    def _notify(self, priority, ctxt, event_type, payload):
        payload = self._serializer.serialize_entity(ctxt, payload)
        # NOTE(Dinesh_Bhor): simulate the kombu serializer
        # this permit to raise an exception if something have not
        # been serialized correctly
        jsonutils.to_primitive(payload)
        # NOTE(Dinesh_Bhor): Try to serialize the context, as the rpc would.
        #                An exception will be raised if something is wrong
        #                with the context.
        self._serializer.serialize_context(ctxt)
        msg = FakeMessage(self.publisher_id, priority, event_type,
                          payload, ctxt)
        NOTIFICATIONS.append(msg)


class FakeVersionedNotifier(FakeNotifier):
    def _notify(self, priority, ctxt, event_type, payload):
        payload = self._serializer.serialize_entity(ctxt, payload)
        VERSIONED_NOTIFICATIONS.append({'publisher_id': self.publisher_id,
                                        'priority': priority,
                                        'event_type': event_type,
                                        'payload': payload})



# Copyright 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from oslo_config import cfg
from oslo_config import types


instance_recovery_group = cfg.OptGroup(
    'instance_failure',
    title='Instance failure recovery options',
    help="Configuration options for instance failure recovery")

host_recovery_group = cfg.OptGroup(
    'host_failure',
    title='Host failure recovery options',
    help="Configuration options for host failure recovery")

customized_recovery_flow_group = cfg.OptGroup(
    'taskflow_driver_recovery_flows',
    title='Customized recovery flow Options',
    help="Configuration options for customizing various failure recovery"
         "workflow tasks.")

taskflow_group = cfg.OptGroup(
    'taskflow',
    title='Taskflow driver options',
    help="Configuration options for taskflow driver")


host_failure_opts = [
    cfg.BoolOpt('evacuate_all_instances',
                default=True,
                help="""
Operators can decide whether all instances or only those instances which
contain metadata key 'HA_Enabled=True' should be allowed for evacuation from
a failed source compute node. When set to True, it will evacuate all instances
from a failed source compute node. First preference will be given to those
instances which contain 'HA_Enabled=True' metadata key, and then it will
evacuate the remaining ones. When set to False, it will evacuate only those
instances which contain 'HA_Enabled=True' metadata key."""),

    cfg.BoolOpt('ignore_instances_in_error_state',
                default=False,
                help="""
Operators can decide whether error instances should be allowed for evacuation
from a failed source compute node or not. When set to True, it will ignore
error instances from evacuation from a failed source compute node. When set to
False, it will evacuate error instances along with other instances from a
failed source compute node."""),

    cfg.BoolOpt("add_reserved_host_to_aggregate",
                default=False,
                help="""
Operators can decide whether reserved_host should be added to aggregate group
of failed compute host. When set to True, reserved host will be added to the
aggregate group of failed compute host. When set to False, the reserved_host
will not be added to the aggregate group of failed compute host."""),
]

instance_failure_options = [
    cfg.BoolOpt('process_all_instances',
                default=False,
                help="""
Operators can decide whether all instances or only those instances which
contain metadata key 'HA_Enabled=True' should be taken into account to
recover from instance failure events. When set to True, it will execute
instance failure recovery actions for an instance irrespective of whether
that particular instance contains metadata key 'HA_Enabled=True' or not.
When set to False, it will only execute instance failure recovery actions
for an instance which contain metadata key 'HA_Enabled=True'."""),
]

taskflow_options = [
    cfg.StrOpt('connection',
               help="""
The SQLAlchemy connection string to use to connect to the taskflow database.
"""),
]

taskflow_driver_recovery_flows = [
    cfg.Opt('host_auto_failure_recovery_tasks',
            type=types.Dict(
                bounds=False,
                value_type=types.List(bounds=True,
                                      item_type=types.String(quotes=True))),
            default={'pre': ['disable_compute_service_task'],
                     'main': ['prepare_HA_enabled_instances_task'],
                     'post': ['evacuate_instances_task']},
            help=("""
This option allows operator to customize tasks to be executed for host failure
auto recovery workflow.

Provide list of strings reflecting to the task classes that should be included
to the host failure recovery workflow. The full classname path of all task
classes should be defined in the 'masakari.task_flow.tasks' of setup.cfg and
these classes may be implemented by OpenStack Masaskari project team, deployer
or third party.

By default below three tasks will be part of this config option:-
1. disable_compute_service_task
2. prepare_HA_enabled_instances_task
3. evacuate_instances_task

The allowed values for this option is comma separated dictionary of object
names in between ``{`` and ``}``.""")),

    cfg.Opt('host_rh_failure_recovery_tasks',
            type=types.Dict(
                bounds=False,
                value_type=types.List(bounds=True,
                                      item_type=types.String(quotes=True))),
            default={'pre': ['disable_compute_service_task'],
                     'main': ['prepare_HA_enabled_instances_task',
                              'evacuate_instances_task'],
                     'post': []},
            help=("""
This option allows operator to customize tasks to be executed for host failure
reserved_host recovery workflow.

Provide list of strings reflecting to the task classes that should be included
to the host failure recovery workflow. The full classname path of all task
classes should be defined in the 'masakari.task_flow.tasks' of setup.cfg and
these classes may be implemented by OpenStack Masaskari project team, deployer
or third party.

By default below three tasks will be part of this config option:-
1. disable_compute_service_task
2. prepare_HA_enabled_instances_task
3. evacuate_instances_task

The allowed values for this option is comma separated dictionary of object
names in between ``{`` and ``}``.""")),

    cfg.Opt('instance_failure_recovery_tasks',
            type=types.Dict(
                bounds=False,
                value_type=types.List(bounds=True,
                                      item_type=types.String(quotes=True))),
            default={'pre': ['stop_instance_task'],
                     'main': ['start_instance_task'],
                     'post': ['confirm_instance_active_task']},
            help=("""
This option allows operator to customize tasks to be executed for instance
failure recovery workflow.

Provide list of strings reflecting to the task classes that should be included
to the instance failure recovery workflow. The full classname path of all task
classes should be defined in the 'masakari.task_flow.tasks' of setup.cfg and
these classes may be implemented by OpenStack Masaskari project team, deployer
or third party.

By default below three tasks will be part of this config option:-
1. stop_instance_task
2. start_instance_task
3. confirm_instance_active_task

The allowed values for this option is comma separated dictionary of object
names in between ``{`` and ``}``.""")),

    cfg.Opt('process_failure_recovery_tasks',
            type=types.Dict(
                bounds=False,
                value_type=types.List(bounds=True,
                                      item_type=types.String(quotes=True))),
            default={'pre': ['disable_compute_node_task'],
                     'main': ['confirm_compute_node_disabled_task'],
                     'post': []},
            help=("""
This option allows operator to customize tasks to be executed for process
failure recovery workflow.

Provide list of strings reflecting to the task classes that should be included
to the process failure recovery workflow. The full classname path of all task
classes should be defined in the 'masakari.task_flow.tasks' of setup.cfg and
these classes may be implemented by OpenStack Masaskari project team, deployer
or third party.

By default below two tasks will be part of this config option:-
1. disable_compute_node_task
2. confirm_compute_node_disabled_task

The allowed values for this option is comma separated dictionary of object
names in between ``{`` and ``}``."""))
]


def register_opts(conf):
    conf.register_group(instance_recovery_group)
    conf.register_group(host_recovery_group)
    conf.register_group(customized_recovery_flow_group)
    conf.register_group(taskflow_group)
    conf.register_opts(instance_failure_options, group=instance_recovery_group)
    conf.register_opts(host_failure_opts, group=host_recovery_group)
    conf.register_opts(taskflow_driver_recovery_flows,
                       group=customized_recovery_flow_group)
    conf.register_opts(taskflow_options, group=taskflow_group)


def list_opts():
    return {
        instance_recovery_group.name: instance_failure_options,
        host_recovery_group.name: host_failure_opts,
        taskflow_group.name: taskflow_options
    }


def customized_recovery_flow_list_opts():
    return {
        customized_recovery_flow_group.name: taskflow_driver_recovery_flows
    }



# Copyright 2016 NTT Data.
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from sqlalchemy import MetaData, Table


def upgrade(migrate_engine):
    meta = MetaData(bind=migrate_engine)

    failover_segments = Table('failover_segments', meta, autoload=True)
    hosts = Table('hosts', meta, autoload=True)

    for table in [failover_segments, hosts]:
        table.c.created_at.alter(nullable=True)



# Copyright (c) 2016 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""The Host API extension."""

from oslo_utils import encodeutils
from oslo_utils import strutils
from six.moves import http_client as http
from webob import exc

from masakari.api.openstack import common
from masakari.api.openstack import extensions
from masakari.api.openstack.ha.schemas import hosts as schema
from masakari.api.openstack import wsgi
from masakari.api import validation
from masakari import exception
from masakari.ha import api as host_api
from masakari.i18n import _
from masakari import objects
from masakari.policies import hosts as host_policies

ALIAS = "os-hosts"


class HostsController(wsgi.Controller):
    """The Host API controller for the OpenStack API."""

    def __init__(self):
        self.api = host_api.HostAPI()

    @extensions.expected_errors((http.BAD_REQUEST, http.FORBIDDEN,
                                 http.NOT_FOUND))
    def index(self, req, segment_id):
        """Returns a list a hosts."""
        context = req.environ['masakari.context']
        context.can(host_policies.HOSTS % 'index')

        try:
            filters = {}
            limit, marker = common.get_limit_and_marker(req)
            sort_keys, sort_dirs = common.get_sort_params(req.params)

            segment = objects.FailoverSegment.get_by_uuid(context,
                                                          segment_id)

            filters['failover_segment_id'] = segment.uuid
            if 'name' in req.params:
                filters['name'] = req.params['name']

            if 'type' in req.params:
                filters['type'] = req.params['type']

            if 'control_attributes' in req.params:
                filters['control_attributes'] = req.params[
                    'control_attributes']

            if 'on_maintenance' in req.params:
                try:
                    filters['on_maintenance'] = strutils.bool_from_string(
                        req.params['on_maintenance'], strict=True)
                except ValueError as ex:
                    msg = _("Invalid value for on_maintenance: "
                            "%s") % encodeutils.exception_to_unicode(ex)
                    raise exc.HTTPBadRequest(explanation=msg)

            if 'reserved' in req.params:
                try:
                    filters['reserved'] = strutils.bool_from_string(
                        req.params['reserved'], strict=True)
                except ValueError as ex:
                    msg = _("Invalid value for reserved: "
                            "%s") % encodeutils.exception_to_unicode(ex)
                    raise exc.HTTPBadRequest(explanation=msg)

            hosts = self.api.get_all(context, filters=filters,
                                     sort_keys=sort_keys, sort_dirs=sort_dirs,
                                     limit=limit, marker=marker)
        except exception.MarkerNotFound as ex:
            raise exc.HTTPBadRequest(explanation=ex.format_message())
        except exception.Invalid as e:
            raise exc.HTTPBadRequest(explanation=e.format_message())
        except exception.FailoverSegmentNotFound as ex:
            raise exc.HTTPNotFound(explanation=ex.format_message())

        return {'hosts': hosts}

    @wsgi.response(http.CREATED)
    @extensions.expected_errors((http.BAD_REQUEST, http.FORBIDDEN,
                                 http.NOT_FOUND, http.CONFLICT))
    @validation.schema(schema.create)
    def create(self, req, segment_id, body):
        """Creates a host."""
        context = req.environ['masakari.context']
        context.can(host_policies.HOSTS % 'create')
        host_data = body.get('host')
        try:
            host = self.api.create_host(context, segment_id, host_data)
        except exception.HypervisorNotFoundByName as e:
            raise exc.HTTPBadRequest(explanation=e.message)
        except exception.FailoverSegmentNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        except exception.HostExists as e:
            raise exc.HTTPConflict(explanation=e.format_message())

        return {'host': host}

    @extensions.expected_errors((http.FORBIDDEN, http.NOT_FOUND))
    def show(self, req, segment_id, id):
        """Shows the details of a host."""
        context = req.environ['masakari.context']
        context.can(host_policies.HOSTS % 'detail')
        try:
            host = self.api.get_host(context, segment_id, id)
        except exception.HostNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        except exception.FailoverSegmentNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        return {'host': host}

    @extensions.expected_errors((http.BAD_REQUEST, http.FORBIDDEN,
                                 http.NOT_FOUND, http.CONFLICT))
    @validation.schema(schema.update)
    def update(self, req, segment_id, id, body):
        """Updates the existing host."""
        context = req.environ['masakari.context']
        context.can(host_policies.HOSTS % 'update')
        host_data = body.get('host')
        try:
            host = self.api.update_host(context, segment_id, id, host_data)
        except exception.HypervisorNotFoundByName as e:
            raise exc.HTTPBadRequest(explanation=e.message)
        except exception.HostNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        except exception.FailoverSegmentNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        except (exception.HostExists, exception.Conflict) as e:
            raise exc.HTTPConflict(explanation=e.format_message())

        return {'host': host}

    @wsgi.response(http.NO_CONTENT)
    @extensions.expected_errors((http.FORBIDDEN, http.NOT_FOUND,
                                 http.CONFLICT))
    def delete(self, req, segment_id, id):
        """Removes a host by id."""
        context = req.environ['masakari.context']
        context.can(host_policies.HOSTS % 'delete')
        try:
            self.api.delete_host(context, segment_id, id)
        except exception.FailoverSegmentNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        except exception.HostNotFound as e:
            raise exc.HTTPNotFound(explanation=e.format_message())
        except exception.Conflict as e:
            raise exc.HTTPConflict(explanation=e.format_message())


class Hosts(extensions.V1APIExtensionBase):
    """Hosts controller"""

    name = "Hosts"
    alias = ALIAS
    version = 1

    def get_resources(self):
        parent = {'member_name': 'segment',
                  'collection_name': 'segments'}
        resources = [
            extensions.ResourceExtension(
                'hosts', HostsController(), parent=parent,
                member_name='host')]

        return resources

    def get_controller_extensions(self):
        return []



#    Copyright 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

import importlib

import eventlet
import mock
from oslo_config import cfg
from oslo_context import context as common_context
from oslo_context import fixture as context_fixture
import six

import masakari
from masakari import context
from masakari import exception
from masakari import test
from masakari import utils

CONF = cfg.CONF


class UTF8TestCase(test.NoDBTestCase):
    def test_none_value(self):
        self.assertIsInstance(utils.utf8(None), type(None))

    def test_bytes_value(self):
        some_value = b"fake data"
        return_value = utils.utf8(some_value)
        # check that type of returned value doesn't changed
        self.assertIsInstance(return_value, type(some_value))
        self.assertEqual(some_value, return_value)

    def test_not_text_type(self):
        return_value = utils.utf8(1)
        self.assertEqual(b"1", return_value)
        self.assertIsInstance(return_value, six.binary_type)

    def test_text_type_with_encoding(self):
        some_value = 'test\u2026config'
        self.assertEqual(some_value, utils.utf8(some_value).decode("utf-8"))


class MonkeyPatchTestCase(test.NoDBTestCase):
    """Unit test for utils.monkey_patch()."""
    def setUp(self):
        super(MonkeyPatchTestCase, self).setUp()
        self.example_package = 'masakari.tests.unit.monkey_patch_example.'
        self.flags(
            monkey_patch=True,
            monkey_patch_modules=[self.example_package + 'example_a' + ':' +
            self.example_package + 'example_decorator'])

    def test_monkey_patch(self):
        utils.monkey_patch()
        masakari.tests.unit.monkey_patch_example.CALLED_FUNCTION = []
        from masakari.tests.unit.monkey_patch_example import example_a
        from masakari.tests.unit.monkey_patch_example import example_b

        self.assertEqual('Example function', example_a.example_function_a())
        exampleA = example_a.ExampleClassA()
        exampleA.example_method()
        ret_a = exampleA.example_method_add(3, 5)
        self.assertEqual(ret_a, 8)

        self.assertEqual('Example function', example_b.example_function_b())
        exampleB = example_b.ExampleClassB()
        exampleB.example_method()
        ret_b = exampleB.example_method_add(3, 5)

        self.assertEqual(ret_b, 8)
        package_a = self.example_package + 'example_a.'
        self.assertIn(package_a + 'example_function_a',
                      masakari.tests.unit.monkey_patch_example.CALLED_FUNCTION)

        self.assertIn(package_a + 'ExampleClassA.example_method',
                      masakari.tests.unit.monkey_patch_example.CALLED_FUNCTION)
        self.assertIn(package_a + 'ExampleClassA.example_method_add',
                      masakari.tests.unit.monkey_patch_example.CALLED_FUNCTION)
        package_b = self.example_package + 'example_b.'
        self.assertNotIn(package_b + 'example_function_b', (
            masakari.tests.unit.monkey_patch_example.CALLED_FUNCTION))
        self.assertNotIn(package_b + 'ExampleClassB.example_method', (
            masakari.tests.unit.monkey_patch_example.CALLED_FUNCTION))
        self.assertNotIn(package_b + 'ExampleClassB.example_method_add', (
            masakari.tests.unit.monkey_patch_example.CALLED_FUNCTION))


class MonkeyPatchDefaultTestCase(test.NoDBTestCase):
    """Unit test for default monkey_patch_modules value."""

    def setUp(self):
        super(MonkeyPatchDefaultTestCase, self).setUp()
        self.flags(
            monkey_patch=True)

    def test_monkey_patch_default_mod(self):
        # monkey_patch_modules is defined to be
        #    <module_to_patch>:<decorator_to_patch_with>
        #  Here we check that both parts of the default values are
        # valid
        for module in CONF.monkey_patch_modules:
            m = module.split(':', 1)
            # Check we can import the module to be patched
            importlib.import_module(m[0])
            # check the decorator is valid
            decorator_name = m[1].rsplit('.', 1)
            decorator_module = importlib.import_module(decorator_name[0])
            getattr(decorator_module, decorator_name[1])


class ExpectedArgsTestCase(test.NoDBTestCase):
    def test_passes(self):
        @utils.expects_func_args('foo', 'baz')
        def dec(f):
            return f

        @dec
        def func(foo, bar, baz="lol"):
            pass

        # Call to ensure nothing errors
        func(None, None)

    def test_raises(self):
        @utils.expects_func_args('foo', 'baz')
        def dec(f):
            return f

        def func(bar, baz):
            pass

        self.assertRaises(TypeError, dec, func)

    def test_var_no_of_args(self):
        @utils.expects_func_args('foo')
        def dec(f):
            return f

        @dec
        def func(bar, *args, **kwargs):
            pass

        # Call to ensure nothing errors
        func(None)

    def test_more_layers(self):
        @utils.expects_func_args('foo', 'baz')
        def dec(f):
            return f

        def dec_2(f):
            def inner_f(*a, **k):
                return f()
            return inner_f

        @dec_2
        def func(bar, baz):
            pass

        self.assertRaises(TypeError, dec, func)


class SpawnNTestCase(test.NoDBTestCase):
    def setUp(self):
        super(SpawnNTestCase, self).setUp()
        self.useFixture(context_fixture.ClearRequestContext())
        self.spawn_name = 'spawn_n'

    def test_spawn_n_no_context(self):
        self.assertIsNone(common_context.get_current())

        def _fake_spawn(func, *args, **kwargs):
            # call the method to ensure no error is raised
            func(*args, **kwargs)
            self.assertEqual('test', args[0])

        def fake(arg):
            pass

        with mock.patch.object(eventlet, self.spawn_name, _fake_spawn):
            getattr(utils, self.spawn_name)(fake, 'test')
        self.assertIsNone(common_context.get_current())

    def test_spawn_n_context(self):
        self.assertIsNone(common_context.get_current())
        ctxt = context.RequestContext('user', 'project')

        def _fake_spawn(func, *args, **kwargs):
            # call the method to ensure no error is raised
            func(*args, **kwargs)
            self.assertEqual(ctxt, args[0])
            self.assertEqual('test', kwargs['kwarg1'])

        def fake(context, kwarg1=None):
            pass

        with mock.patch.object(eventlet, self.spawn_name, _fake_spawn):
            getattr(utils, self.spawn_name)(fake, ctxt, kwarg1='test')
        self.assertEqual(ctxt, common_context.get_current())

    def test_spawn_n_context_different_from_passed(self):
        self.assertIsNone(common_context.get_current())
        ctxt = context.RequestContext('user', 'project')
        ctxt_passed = context.RequestContext('user', 'project',
                overwrite=False)
        self.assertEqual(ctxt, common_context.get_current())

        def _fake_spawn(func, *args, **kwargs):
            # call the method to ensure no error is raised
            func(*args, **kwargs)
            self.assertEqual(ctxt_passed, args[0])
            self.assertEqual('test', kwargs['kwarg1'])

        def fake(context, kwarg1=None):
            pass

        with mock.patch.object(eventlet, self.spawn_name, _fake_spawn):
            getattr(utils, self.spawn_name)(fake, ctxt_passed, kwarg1='test')
        self.assertEqual(ctxt, common_context.get_current())


class SpawnTestCase(SpawnNTestCase):
    def setUp(self):
        super(SpawnTestCase, self).setUp()
        self.spawn_name = 'spawn'


class ValidateIntegerTestCase(test.NoDBTestCase):
    def test_exception_converted(self):
        self.assertRaises(exception.InvalidInput,
                          utils.validate_integer,
                          "im-not-an-int", "not-an-int")
        self.assertRaises(exception.InvalidInput,
                          utils.validate_integer,
                          3.14, "Pie")
        self.assertRaises(exception.InvalidInput,
                          utils.validate_integer,
                          "299", "Sparta no-show",
                          min_value=300, max_value=300)
        self.assertRaises(exception.InvalidInput,
                          utils.validate_integer,
                          55, "doing 55 in a 54",
                          max_value=54)
        self.assertRaises(exception.InvalidInput,
                          utils.validate_integer,
                          six.unichr(129), "UnicodeError",
                          max_value=1000)



# Copyright 2016 NTT DATA
# All Rights Reserved.

#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""
Unit Tests for process failure TaskFlow
"""

import mock

from masakari.compute import nova
from masakari import context
from masakari.engine.drivers.taskflow import process_failure
from masakari import exception
from masakari import test
from masakari.tests.unit import fakes


class ProcessFailureTestCase(test.TestCase):

    def setUp(self):
        super(ProcessFailureTestCase, self).setUp()
        self.ctxt = context.get_admin_context()
        self.process_name = "nova-compute"
        self.service_host = "fake-host"
        self.novaclient = nova.API()
        self.fake_client = fakes.FakeNovaClient()
        # overriding 'wait_period_after_service_update' to 2 seconds
        # to reduce the wait period.
        self.override_config('wait_period_after_service_update', 2)

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_compute_process_failure_flow(self, _mock_notify,
                                          _mock_novaclient):
        _mock_novaclient.return_value = self.fake_client

        # create test data
        self.fake_client.services.create("1", host=self.service_host,
                                         binary="nova-compute",
                                         status="enabled")

        # test DisableComputeNodeTask
        task = process_failure.DisableComputeNodeTask(self.ctxt,
                                                      self.novaclient)
        task.execute(self.process_name, self.service_host)

        # test ConfirmComputeNodeDisabledTask
        task = process_failure.ConfirmComputeNodeDisabledTask(self.ctxt,
                                                              self.novaclient)
        task.execute(self.process_name, self.service_host)

        # verify service is disabled
        self.assertTrue(self.novaclient.is_service_down(self.ctxt,
                                                        self.service_host,
                                                        self.process_name))

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0),
            mock.call("Confirming compute service is disabled on host: "
                      "'fake-host'"),
            mock.call("Confirmed compute service is disabled on host: "
                      "'fake-host'", 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_compute_process_failure_flow_disabled_process(self,
                                                           _mock_notify,
                                                           _mock_novaclient):
        _mock_novaclient.return_value = self.fake_client

        # create test data
        self.fake_client.services.create("1", host=self.service_host,
                                         binary="nova-compute",
                                         status="disabled")

        # test DisableComputeNodeTask
        task = process_failure.DisableComputeNodeTask(self.ctxt,
                                                      self.novaclient)
        with mock.patch.object(
            self.novaclient,
            'enable_disable_service') as mock_enable_disabled:
            task.execute(self.process_name, self.service_host)

        # ensure that enable_disable_service method is not called
        self.assertEqual(0, mock_enable_disabled.call_count)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call('Skipping recovery for process nova-compute as it is '
                      'already disabled', 1.0)
        ])

    @mock.patch('masakari.compute.nova.novaclient')
    @mock.patch('masakari.engine.drivers.taskflow.base.MasakariTask.'
                'update_details')
    def test_compute_process_failure_flow_compute_service_disabled_failed(
        self, _mock_notify, _mock_novaclient):
        _mock_novaclient.return_value = self.fake_client

        # create test data
        self.fake_client.services.create("1", host=self.service_host,
                                         binary="nova-compute",
                                         status="enabled")

        def fake_is_service_down(context, host_name, binary):
            # assume that service is not disabled
            return False

        # test DisableComputeNodeTask
        task = process_failure.DisableComputeNodeTask(self.ctxt,
                                                      self.novaclient)
        task.execute(self.process_name, self.service_host)

        with mock.patch.object(self.novaclient, 'is_service_down',
                               fake_is_service_down):
            # test ConfirmComputeNodeDisabledTask
            task = process_failure.ConfirmComputeNodeDisabledTask(
                self.ctxt, self.novaclient)
            self.assertRaises(exception.ProcessRecoveryFailureException,
                              task.execute, self.process_name,
                              self.service_host)

        # verify progress details
        _mock_notify.assert_has_calls([
            mock.call("Disabling compute service on host: 'fake-host'"),
            mock.call("Disabled compute service on host: 'fake-host'", 1.0),
            mock.call("Confirming compute service is disabled on host: "
                      "'fake-host'"),
            mock.call('Failed to disable service nova-compute', 1.0)
        ])



# Copyright 2016 NTT DATA
#
# Licensed under the Apache License, Version 2.0 (the "License"); you may
# not use this file except in compliance with the License. You may obtain
# a copy of the License at
#
#    http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
# WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
# License for the specific language governing permissions and limitations
# under the License.

import os
import tempfile

import fixtures
from oslo_config import cfg

import masakari.conf.api
from masakari import test


class ConfTest(test.NoDBTestCase):
    """This is a test and pattern for parsing tricky options."""

    class TestConfigOpts(cfg.ConfigOpts):
        def __call__(self, args=None, default_config_files=None):
            if default_config_files is None:
                default_config_files = []
            return cfg.ConfigOpts.__call__(
                self,
                args=args,
                prog='test',
                version='1.0',
                usage='%(prog)s FOO BAR',
                default_config_files=default_config_files,
                validate_default_values=True)

    def setUp(self):
        super(ConfTest, self).setUp()
        self.useFixture(fixtures.NestedTempfile())
        self.conf = self.TestConfigOpts()
        self.tempdirs = []

    def create_tempfiles(self, files, ext='.conf'):
        tempfiles = []
        for (basename, contents) in files:
            if not os.path.isabs(basename):
                (fd, path) = tempfile.mkstemp(prefix=basename, suffix=ext)
            else:
                path = basename + ext
                fd = os.open(path, os.O_CREAT | os.O_WRONLY)
            tempfiles.append(path)
            try:
                os.write(fd, contents.encode('utf-8'))
            finally:
                os.close(fd)
        return tempfiles

    def test_reserved_huge_page(self):
        masakari.conf.api.register_opts(self.conf)

        paths = self.create_tempfiles(
            [('1',
              '[DEFAULT]\n'
              'osapi_max_limit = 1000\n')])
        self.conf(['--config-file', paths[0]])
        # NOTE(Dinesh_Bhor): In oslo.config if you specify a parameter
        # incorrectly, it silently drops it from the conf. Which means
        # the attr doesn't exist at all. The first attr test here is
        # for an unrelated boolean option that is using defaults (so
        # will always work. It's a basic control that *anything* is working.
        self.assertTrue(hasattr(self.conf, 'osapi_max_limit'))
        self.assertTrue(hasattr(self.conf, 'use_forwarded_for'))

        # NOTE(Dinesh_Bhor): Yes, this actually parses as an array holding
        # a dict.
        actual = 1000
        self.assertEqual(actual, self.conf.osapi_max_limit)



# Copyright 2017 NTT DATA
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

""" VM libvirt events

These are the events which needs to be processed by masakari in case of
instance recovery failure.
"""

INSTANCE_EVENTS = {
    # Add QEMU guest agent events here.
    'QEMU_GUEST_AGENT_ERROR': ['STOPPED_FAILED'],
    # Add more events and vir_domain_events here.
    'LIFECYCLE': ['STOPPED_FAILED'],
    'IO_ERROR': ['IO_ERROR_REPORT']
}


def is_valid_event(payload):
    vir_domain_event_list = INSTANCE_EVENTS.get(payload.get('event'))
    if vir_domain_event_list and payload.get(
            'vir_domain_event') in vir_domain_event_list:
        return True
    return False



# Copyright (c) 2016 NTT Data
# All Rights Reserved.
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

from six.moves import http_client
import webob.exc

from masakari.api.openstack import extensions
from masakari.api.openstack.ha import versions
from masakari.api.openstack.ha.views import versions as views_versions
from masakari.api.openstack import wsgi


ALIAS = "versions"


class VersionsController(wsgi.Controller):
    @extensions.expected_errors(http_client.NOT_FOUND)
    def show(self, req, id='v1'):
        builder = views_versions.get_view_builder(req)
        if id not in versions.VERSIONS:
            raise webob.exc.HTTPNotFound()

        return builder.build_version(versions.VERSIONS[id])


class Versions(extensions.V1APIExtensionBase):
    """API Version information."""

    name = "Versions"
    alias = ALIAS
    version = 1

    def get_resources(self):
        resources = [
            extensions.ResourceExtension(ALIAS, VersionsController(),
                                         custom_routes_fn=self.version_map)]
        return resources

    def get_controller_extensions(self):
        return []

    def version_map(self, mapper, wsgi_resource):
        mapper.connect("versions", "/",
                       controller=wsgi_resource,
                       action='show', conditions={"method": ['GET']})
        mapper.redirect("", "/")



# Copyright (c) 2016 NTT DATA
#
#    Licensed under the Apache License, Version 2.0 (the "License"); you may
#    not use this file except in compliance with the License. You may obtain
#    a copy of the License at
#
#         http://www.apache.org/licenses/LICENSE-2.0
#
#    Unless required by applicable law or agreed to in writing, software
#    distributed under the License is distributed on an "AS IS" BASIS, WITHOUT
#    WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied. See the
#    License for the specific language governing permissions and limitations
#    under the License.

"""Utility methods for working with WSGI servers."""

from __future__ import print_function

import os.path
import socket
import ssl
import sys

import eventlet
import eventlet.wsgi
import greenlet
from oslo_log import log as logging
from oslo_service import service
from oslo_utils import excutils
from paste import deploy
import routes.middleware
import webob.dec
import webob.exc

import masakari.conf
from masakari import exception
from masakari.i18n import _
from masakari import utils

CONF = masakari.conf.CONF

LOG = logging.getLogger(__name__)


class Server(service.ServiceBase):
    """Server class to manage a WSGI server, serving a WSGI application."""

    default_pool_size = CONF.wsgi.default_pool_size

    def __init__(self, name, app, host='0.0.0.0', port=0, pool_size=None,
                 protocol=eventlet.wsgi.HttpProtocol, backlog=128,
                 use_ssl=False, max_url_len=None):
        """Initialize, but do not start, a WSGI server.

        :param name: Pretty name for logging.
        :param app: The WSGI application to serve.
        :param host: IP address to serve the application.
        :param port: Port number to server the application.
        :param pool_size: Maximum number of eventlets to spawn concurrently.
        :param backlog: Maximum number of queued connections.
        :param max_url_len: Maximum length of permitted URLs.
        :returns: None
        :raises: masakari.exception.InvalidInput
        """
        # Allow operators to customize http requests max header line size.
        eventlet.wsgi.MAX_HEADER_LINE = CONF.wsgi.max_header_line
        self.name = name
        self.app = app
        self._server = None
        self._protocol = protocol
        self.pool_size = pool_size or self.default_pool_size
        self._pool = eventlet.GreenPool(self.pool_size)
        self._logger = logging.getLogger("masakari.%s.wsgi.server" % self.name)
        self._use_ssl = use_ssl
        self._max_url_len = max_url_len

        self.client_socket_timeout = CONF.wsgi.client_socket_timeout or None

        if backlog < 1:
            raise exception.InvalidInput(
                reason=_('The backlog must be more than 0'))

        bind_addr = (host, port)
        try:
            info = socket.getaddrinfo(bind_addr[0],
                                      bind_addr[1],
                                      socket.AF_UNSPEC,
                                      socket.SOCK_STREAM)[0]
            family = info[0]
            bind_addr = info[-1]
        except Exception:
            family = socket.AF_INET

        try:
            self._socket = eventlet.listen(bind_addr, family, backlog=backlog)
        except EnvironmentError:
            LOG.error("Could not bind to %(host)s:%(port)d",
                      {'host': host, 'port': port})
            raise

        (self.host, self.port) = self._socket.getsockname()[0:2]
        LOG.info("%(name)s listening on %(host)s:%(port)d",
                 {'name': self.name, 'host': self.host, 'port': self.port})

    def start(self):
        """Start serving a WSGI application.

        :returns: None
        """
        # The server socket object will be closed after server exits,
        # but the underlying file descriptor will remain open, and will
        # give bad file descriptor error. So duplicating the socket object,
        # to keep file descriptor usable.

        dup_socket = self._socket.dup()
        dup_socket.setsockopt(socket.SOL_SOCKET,
                              socket.SO_REUSEADDR, 1)
        # sockets can hang around forever without keepalive
        dup_socket.setsockopt(socket.SOL_SOCKET,
                              socket.SO_KEEPALIVE, 1)

        # This option isn't available in the OS X version of eventlet
        if hasattr(socket, 'TCP_KEEPIDLE'):
            dup_socket.setsockopt(socket.IPPROTO_TCP,
                                  socket.TCP_KEEPIDLE,
                                  CONF.wsgi.tcp_keepidle)

        if self._use_ssl:
            try:
                ca_file = CONF.wsgi.ssl_ca_file
                cert_file = CONF.wsgi.ssl_cert_file
                key_file = CONF.wsgi.ssl_key_file

                if cert_file and not os.path.exists(cert_file):
                    raise RuntimeError(
                        _("Unable to find cert_file : %s") % cert_file)

                if ca_file and not os.path.exists(ca_file):
                    raise RuntimeError(
                        _("Unable to find ca_file : %s") % ca_file)

                if key_file and not os.path.exists(key_file):
                    raise RuntimeError(
                        _("Unable to find key_file : %s") % key_file)

                if self._use_ssl and (not cert_file or not key_file):
                    raise RuntimeError(
                        _("When running server in SSL mode, you must "
                          "specify both a cert_file and key_file "
                          "option value in your configuration file"))
                ssl_kwargs = {
                    'server_side': True,
                    'certfile': cert_file,
                    'keyfile': key_file,
                    'cert_reqs': ssl.CERT_NONE,
                }

                if CONF.wsgi.ssl_ca_file:
                    ssl_kwargs['ca_certs'] = ca_file
                    ssl_kwargs['cert_reqs'] = ssl.CERT_REQUIRED

                dup_socket = eventlet.wrap_ssl(dup_socket,
                                               **ssl_kwargs)
            except Exception:
                with excutils.save_and_reraise_exception():
                    LOG.error("Failed to start %(name)s on %(host)s"
                              ":%(port)d with SSL support",
                              {'name': self.name, 'host': self.host,
                               'port': self.port})

        wsgi_kwargs = {
            'func': eventlet.wsgi.server,
            'sock': dup_socket,
            'site': self.app,
            'protocol': self._protocol,
            'custom_pool': self._pool,
            'log': self._logger,
            'log_format': CONF.wsgi.wsgi_log_format,
            'debug': False,
            'keepalive': CONF.wsgi.keep_alive,
            'socket_timeout': self.client_socket_timeout
            }

        if self._max_url_len:
            wsgi_kwargs['url_length_limit'] = self._max_url_len

        self._server = utils.spawn(**wsgi_kwargs)

    def reset(self):
        """Reset server greenpool size to default.

        :returns: None

        """
        self._pool.resize(self.pool_size)

    def stop(self):
        """Stop this server.

        This is not a very nice action, as currently the method by which a
        server is stopped is by killing its eventlet.

        :returns: None

        """
        LOG.info("Stopping WSGI server.")

        if self._server is not None:
            # Resize pool to stop new requests from being processed
            self._pool.resize(0)
            self._server.kill()

    def wait(self):
        """Block, until the server has stopped.

        Waits on the server's eventlet to finish, then returns.

        :returns: None

        """
        try:
            if self._server is not None:
                self._pool.waitall()
                self._server.wait()
        except greenlet.GreenletExit:
            LOG.info("WSGI server has stopped.")


class Request(webob.Request):
    def __init__(self, environ, *args, **kwargs):
        if CONF.wsgi.secure_proxy_ssl_header:
            scheme = environ.get(CONF.wsgi.secure_proxy_ssl_header)
            if scheme:
                environ['wsgi.url_scheme'] = scheme
        super(Request, self).__init__(environ, *args, **kwargs)


class Application(object):
    """Base WSGI application wrapper. Subclasses need to implement __call__."""

    @classmethod
    def factory(cls, global_config, **local_config):
        """Used for paste app factories in paste.deploy config files.

        Any local configuration (that is, values under the [app:APPNAME]
        section of the paste config) will be passed into the `__init__` method
        as kwargs.

        A hypothetical configuration would look like:

            [app:wadl]
            latest_version = 1.3
            paste.app_factory = masakari.api.fancy_api:Wadl.factory

        which would result in a call to the `Wadl` class as

            import masakari.api.fancy_api
            fancy_api.Wadl(latest_version='1.3')

        You could of course re-implement the `factory` method in subclasses,
        but using the kwarg passing it shouldn't be necessary.

        """
        return cls(**local_config)

    def __call__(self, environ, start_response):
        r"""Subclasses will probably want to implement __call__ like this:

        @webob.dec.wsgify(RequestClass=Request)
        def __call__(self, req):
          # Any of the following objects work as responses:

          # Option 1: simple string
          res = 'message\n'

          # Option 2: a nicely formatted HTTP exception page
          res = exc.HTTPForbidden(explanation='Nice try')

          # Option 3: a webob Response object (in case you need to play with
          # headers, or you want to be treated like an iterable, or ...)
          res = Response()
          res.app_iter = open('somefile')

          # Option 4: any wsgi app to be run next
          res = self.application

          # Option 5: you can get a Response object for a wsgi app, too, to
          # play with headers etc
          res = req.get_response(self.application)

          # You can then just return your response...
          return res
          # ... or set req.response and return None.
          req.response = res

        See the end of http://pythonpaste.org/webob/modules/dec.html
        for more info.

        """
        raise NotImplementedError(_('You must implement __call__'))


class Middleware(Application):
    """Base WSGI middleware.

    These classes require an application to be
    initialized that will be called next.  By default the middleware will
    simply call its wrapped app, or you can override __call__ to customize its
    behavior.

    """

    @classmethod
    def factory(cls, global_config, **local_config):
        """Used for paste app factories in paste.deploy config files.

        Any local configuration (that is, values under the [filter:APPNAME]
        section of the paste config) will be passed into the `__init__` method
        as kwargs.

        A hypothetical configuration would look like:

            [filter:analytics]
            redis_host = 127.0.0.1
            paste.filter_factory = masakari.api.analytics:Analytics.factory

        which would result in a call to the `Analytics` class as

            import masakari.api.analytics
            analytics.Analytics(app_from_paste, redis_host='127.0.0.1')

        You could of course re-implement the `factory` method in subclasses,
        but using the kwarg passing it shouldn't be necessary.

        """
        def _factory(app):
            return cls(app, **local_config)
        return _factory

    def __init__(self, application):
        self.application = application

    def process_request(self, req):
        """Called on each request.

        If this returns None, the next application down the stack will be
        executed. If it returns a response then that response will be returned
        and execution will stop here.

        """
        return None

    def process_response(self, response):
        """Do whatever you'd like to the response."""
        return response

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, req):
        response = self.process_request(req)
        if response:
            return response
        response = req.get_response(self.application)
        return self.process_response(response)


class Debug(Middleware):
    """Helper class for debugging a WSGI application.

    Can be inserted into any WSGI application chain to get information
    about the request and response.

    """

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, req):
        print(('*' * 40) + ' REQUEST ENVIRON')
        for key, value in req.environ.items():
            print(key, '=', value)
        print()
        resp = req.get_response(self.application)

        print(('*' * 40) + ' RESPONSE HEADERS')
        for (key, value) in resp.headers.items():
            print(key, '=', value)
        print()

        resp.app_iter = self.print_generator(resp.app_iter)

        return resp

    @staticmethod
    def print_generator(app_iter):
        """Iterator that prints the contents of a wrapper string."""
        print(('*' * 40) + ' BODY')
        for part in app_iter:
            sys.stdout.write(part)
            sys.stdout.flush()
            yield part
        print()


class Router(object):
    """WSGI middleware that maps incoming requests to WSGI apps."""

    def __init__(self, mapper):
        """Create a router for the given routes.Mapper.

        Each route in `mapper` must specify a 'controller', which is a
        WSGI app to call.  You'll probably want to specify an 'action' as
        well and have your controller be an object that can route
        the request to the action-specific method.

        Examples:
          mapper = routes.Mapper()
          sc = ServerController()

          # Explicit mapping of one route to a controller+action
          mapper.connect(None, '/svrlist', controller=sc, action='list')

          # Actions are all implicitly defined
          mapper.resource('server', 'servers', controller=sc)

          # Pointing to an arbitrary WSGI app.  You can specify the
          # {path_info:.*} parameter so the target app can be handed just that
          # section of the URL.
          mapper.connect(None, '/v1.0/{path_info:.*}', controller=BlogApp())

        """
        self.map = mapper
        self._router = routes.middleware.RoutesMiddleware(self._dispatch,
                                                          self.map)

    @webob.dec.wsgify(RequestClass=Request)
    def __call__(self, req):
        """Route the incoming request to a controller based on self.map.

        If no match, return a 404.

        """
        return self._router

    @staticmethod
    @webob.dec.wsgify(RequestClass=Request)
    def _dispatch(req):
        """Dispatch the request to the appropriate controller.

        Called by self._router after matching the incoming request to a route
        and putting the information into req.environ.  Either returns 404
        or the routed WSGI app's response.

        """
        match = req.environ['wsgiorg.routing_args'][1]
        if not match:
            return webob.exc.HTTPNotFound()
        app = match['controller']
        return app


class Loader(object):
    """Used to load WSGI applications from paste configurations."""

    def __init__(self, config_path=None):
        """Initialize the loader, and attempt to find the config.

        :param config_path: Full or relative path to the paste config.
        :returns: None

        """
        self.config_path = None

        config_path = config_path or CONF.wsgi.api_paste_config
        if not os.path.isabs(config_path):
            self.config_path = CONF.find_file(config_path)
        elif os.path.exists(config_path):
            self.config_path = config_path

        if not self.config_path:
            raise exception.ConfigNotFound(path=config_path)

    def load_app(self, name):
        """Return the paste URLMap wrapped WSGI application.

        :param name: Name of the application to load.
        :returns: Paste URLMap object wrapping the requested application.
        :raises: `masakari.exception.PasteAppNotFound`

        """
        try:
            LOG.debug("Loading app %(name)s from %(path)s",
                      {'name': name, 'path': self.config_path})
            return deploy.loadapp("config:%s" % self.config_path, name=name)
        except LookupError:
            LOG.exception("Couldn't lookup app: %s", name)
            raise exception.PasteAppNotFound(name=name, path=self.config_path)
